#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright (c) 2024 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

import os
import re
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, get_topogen

pytestmark = [pytest.mark.bgpd]


def setup_module(mod):
    topodef = {"s1": ("r1", "r2", "r3"), "s2": ("r1", "r4")}
    tgen = Topogen(topodef, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_frr_config(os.path.join(CWD, "{}/frr.conf".format(rname)))

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_remote_as_auto():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"]
    r2 = tgen.gears["r2"]
    r3 = tgen.gears["r3"]
    r4 = tgen.gears["r4"]

    def _bgp_converge():
        output = json.loads(r1.vtysh_cmd("show bgp ipv4 unicast summary json"))
        expected = {
            "peers": {
                "r1-eth1": {
                    "hostname": "r4",
                    "remoteAs": 65004,
                    "localAs": 65001,
                    "state": "Established",
                },
                "192.168.1.2": {
                    "hostname": "r2",
                    "remoteAs": 65001,
                    "localAs": 65001,
                    "state": "Established",
                },
                "192.168.1.3": {
                    "hostname": "r3",
                    "remoteAs": 65003,
                    "localAs": 65001,
                    "state": "Established",
                },
            }
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_converge,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "Can't see automatic iBGP/eBGP peerings"

    def _bgp_converge_internal():
        output = json.loads(r2.vtysh_cmd("show bgp ipv4 unicast 10.0.0.1/32 json"))
        expected = {
            "paths": [
                {
                    "aspath": {
                        "string": "Local",
                    },
                    "valid": True,
                    "peer": {
                        "hostname": "r1",
                        "type": "internal",
                    },
                }
            ]
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_converge_internal,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "Can't see automatic iBGP peering"

    def _bgp_converge_external():
        output = json.loads(r3.vtysh_cmd("show bgp ipv4 unicast 10.0.0.1/32 json"))
        expected = {
            "paths": [
                {
                    "aspath": {
                        "string": "65001",
                    },
                    "valid": True,
                    "peer": {
                        "hostname": "r1",
                        "type": "external",
                    },
                }
            ]
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_converge_external,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "Can't see automatic eBGP peering"

    def _bgp_converge_external_unnumbered():
        output = json.loads(r4.vtysh_cmd("show bgp ipv4 unicast 10.0.0.1/32 json"))
        expected = {
            "paths": [
                {
                    "aspath": {
                        "string": "65001",
                    },
                    "valid": True,
                    "peer": {
                        "hostname": "r1",
                        "type": "external",
                    },
                }
            ]
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_converge_external_unnumbered,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "Can't see automatic unnumbered eBGP peering"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: GPL-2.0-or-later
#
# September 2 2021, Christian Hopps <chopps@labn.net>
#
# Copyright 2021, LabN Consulting, L.L.C.
#
"""The main function for standalone operation."""
import argparse
import asyncio
import logging
import logging.config
import os
import subprocess
import sys

from . import cli
from . import parser
from .args import add_launch_args
from .base import get_event_loop
from .cleanup import cleanup_previous
from .cleanup import is_running_in_rundir
from .compat import PytestConfig


logger = None


async def forever():
    while True:
        await asyncio.sleep(3600)


async def run_and_wait(args, unet):
    tasks = []

    if not args.topology_only:
        # add the cmd.wait()s returned from unet.run()
        tasks += await unet.run()

    if sys.stdin.isatty() and not args.no_cli:
        # Run an interactive CLI
        task = cli.async_cli(unet)
    else:
        if args.no_wait:
            logger.info("Waiting for all node cmd to complete")
            task = asyncio.gather(*tasks, return_exceptions=True)
        else:
            logger.info("Waiting on signal to exit")
            task = asyncio.create_task(forever())
            task = asyncio.gather(task, *tasks, return_exceptions=True)

    try:
        await task
    finally:
        # Basically we are canceling tasks from unet.run() which are just async calls to
        # node.cmd_p.wait() so we've stopped waiting for them to complete, but not
        # actually canceld/killed the cmd_p process.
        for task in tasks:
            task.cancel()


async def async_main(args, config):
    status = 3

    # Setup the namespaces and network addressing.

    unet = await parser.async_build_topology(
        config, rundir=args.rundir, args=args, pytestconfig=PytestConfig(args)
    )
    logger.info("Topology up: rundir: %s", unet.rundir)

    try:
        status = await run_and_wait(args, unet)
    except KeyboardInterrupt:
        logger.info("Exiting, received KeyboardInterrupt in async_main")
    except asyncio.CancelledError as ex:
        logger.info("task canceled error: %s cleaning up", ex)
    except Exception as error:
        logger.info("Exiting, unexpected exception %s", error, exc_info=True)
    else:
        logger.info("Exiting normally")

    logger.debug("main: async deleting")
    try:
        await unet.async_delete()
    except KeyboardInterrupt:
        status = 2
        logger.warning("Received KeyboardInterrupt while cleaning up.")
    except Exception as error:
        status = 2
        logger.info("Deleting, unexpected exception %s", error, exc_info=True)
    return status


def main(*args):
    ap = argparse.ArgumentParser(args)
    cap = ap.add_argument_group(title="Config", description="config related options")

    cap.add_argument("-c", "--config", help="config file (yaml, toml, json, ...)")
    cap.add_argument(
        "-d", "--rundir", help="runtime directory for tempfiles, logs, etc"
    )
    cap.add_argument(
        "--kinds-config",
        help="kinds config file, overrides default search (yaml, toml, json, ...)",
    )
    cap.add_argument(
        "--project-root", help="directory to stop searching for kinds config at"
    )

    rap = ap.add_argument_group(title="Runtime", description="runtime related options")
    add_launch_args(rap.add_argument)

    # Move to munet.args?
    rap.add_argument(
        "-C",
        "--cleanup",
        action="store_true",
        help="Remove the entire rundir (not just node subdirs) prior to running.",
    )
    # Move to munet.args?
    rap.add_argument(
        "--topology-only",
        action="store_true",
        help="Do not run any node commands",
    )
    rap.add_argument(
        "--validate-only",
        action="store_true",
        help="Validate the config against the schema definition",
    )
    rap.add_argument("--unshare-inline", action="store_true", help=argparse.SUPPRESS)

    rap.add_argument("-v", "--verbose", action="store_true", help="be verbose")
    rap.add_argument(
        "-V", "--version", action="store_true", help="print the verison number and exit"
    )

    eap = ap.add_argument_group(title="Uncommon", description="uncommonly used options")
    eap.add_argument("--log-config", help="logging config file (yaml, toml, json, ...)")
    eap.add_argument(
        "--kill",
        action="store_true",
        help="Kill previous running processes using same rundir and exit",
    )
    eap.add_argument("--no-kill", action="store_true", help=argparse.SUPPRESS)
    eap.add_argument(
        "--no-cli", action="store_true", help="Do not run the interactive CLI"
    )
    eap.add_argument("--no-wait", action="store_true", help="Exit after commands")

    args = ap.parse_args()

    if args.version:
        from importlib import metadata  # pylint: disable=C0415

        print(metadata.version("munet"))
        sys.exit(0)

    rundir = args.rundir if args.rundir else "/tmp/munet"
    rundir = os.path.abspath(rundir)
    args.rundir = rundir

    if args.kill:
        logging.info("Killing any previous run using rundir: {rundir}")
        cleanup_previous(args.rundir)
    elif is_running_in_rundir(args.rundir):
        logging.fatal(
            "Munet processes using rundir: %s, use `--kill` to cleanup first", rundir
        )
        return 1

    if args.cleanup:
        if os.path.exists(rundir):
            if not os.path.exists(f"{rundir}/config.json"):
                logging.critical(
                    'unsafe: won\'t clean up rundir "%s" as '
                    "previous config.json not present",
                    rundir,
                )
                sys.exit(1)
            else:
                subprocess.run(["/usr/bin/rm", "-rf", rundir], check=True)

    if args.kill:
        return 0

    subprocess.run(f"mkdir -p {rundir} && chmod 755 {rundir}", check=True, shell=True)
    os.environ["MUNET_RUNDIR"] = rundir

    parser.setup_logging(args)

    global logger  # pylint: disable=W0603
    logger = logging.getLogger("munet")

    config = parser.get_config(args.config)
    logger.info("Loaded config from %s", config["config_pathname"])
    if not config["topology"]["nodes"]:
        logger.critical("No nodes defined in config file")
        return 1

    loop = None
    status = 4
    try:
        parser.validate_config(config, logger, args)
        if args.validate_only:
            return 0
        # Executes the cmd for each node.
        loop = get_event_loop()
        status = loop.run_until_complete(async_main(args, config))
    except KeyboardInterrupt:
        logger.info("Exiting, received KeyboardInterrupt in main")
    except Exception as error:
        logger.info("Exiting, unexpected exception %s", error, exc_info=True)
    finally:
        if loop:
            loop.close()

    return status


if __name__ == "__main__":
    exit_status = main()
    sys.exit(exit_status)



#!/usr/bin/env python3
# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: GPL-2.0-or-later
#
# December 22 2021, Christian Hopps <chopps@labn.net>
#
# Copyright 2021-2022, LabN Consulting, L.L.C.
#

import argparse
import asyncio
import errno
import logging
import socket
import struct
import sys
from asyncio import Event, Lock
from ipaddress import ip_address as ip

FMT_APIMSGHDR = ">BBHL"
FMT_APIMSGHDR_SIZE = struct.calcsize(FMT_APIMSGHDR)

FMT_LSA_FILTER = ">HBB"  # + plus x"I" areas
LSAF_ORIGIN_NON_SELF = 0
LSAF_ORIGIN_SELF = 1
LSAF_ORIGIN_ANY = 2

FMT_LSA_HEADER = ">HBBIILHH"
FMT_LSA_HEADER_SIZE = struct.calcsize(FMT_LSA_HEADER)

# ------------------------
# Messages to OSPF daemon.
# ------------------------

MSG_REGISTER_OPAQUETYPE = 1
MSG_UNREGISTER_OPAQUETYPE = 2
MSG_REGISTER_EVENT = 3
MSG_SYNC_LSDB = 4
MSG_ORIGINATE_REQUEST = 5
MSG_DELETE_REQUEST = 6
MSG_SYNC_REACHABLE = 7
MSG_SYNC_ISM = 8
MSG_SYNC_NSM = 9
MSG_SYNC_ROUTER_ID = 19

smsg_info = {
    MSG_REGISTER_OPAQUETYPE: ("REGISTER_OPAQUETYPE", "BBxx"),
    MSG_UNREGISTER_OPAQUETYPE: ("UNREGISTER_OPAQUETYPE", "BBxx"),
    MSG_REGISTER_EVENT: ("REGISTER_EVENT", FMT_LSA_FILTER),
    MSG_SYNC_LSDB: ("SYNC_LSDB", FMT_LSA_FILTER),
    MSG_ORIGINATE_REQUEST: ("ORIGINATE_REQUEST", ">II" + FMT_LSA_HEADER[1:]),
    MSG_DELETE_REQUEST: ("DELETE_REQUEST", ">IBBxBL"),
    MSG_SYNC_REACHABLE: ("MSG_SYNC_REACHABLE", ""),
    MSG_SYNC_ISM: ("MSG_SYNC_ISM", ""),
    MSG_SYNC_NSM: ("MSG_SYNC_NSM", ""),
    MSG_SYNC_ROUTER_ID: ("MSG_SYNC_ROUTER_ID", ""),
}

# OSPF API MSG Delete Flag.
OSPF_API_DEL_ZERO_LEN_LSA = 0x01  # send withdrawal with no LSA data

# --------------------------
# Messages from OSPF daemon.
# --------------------------

MSG_REPLY = 10
MSG_READY_NOTIFY = 11
MSG_LSA_UPDATE_NOTIFY = 12
MSG_LSA_DELETE_NOTIFY = 13
MSG_NEW_IF = 14
MSG_DEL_IF = 15
MSG_ISM_CHANGE = 16
MSG_NSM_CHANGE = 17
MSG_REACHABLE_CHANGE = 18
MSG_ROUTER_ID_CHANGE = 20

amsg_info = {
    MSG_REPLY: ("REPLY", "bxxx"),
    MSG_READY_NOTIFY: ("READY_NOTIFY", ">BBxxI"),
    MSG_LSA_UPDATE_NOTIFY: ("LSA_UPDATE_NOTIFY", ">IIBxxx" + FMT_LSA_HEADER[1:]),
    MSG_LSA_DELETE_NOTIFY: ("LSA_DELETE_NOTIFY", ">IIBxxx" + FMT_LSA_HEADER[1:]),
    MSG_NEW_IF: ("NEW_IF", ">II"),
    MSG_DEL_IF: ("DEL_IF", ">I"),
    MSG_ISM_CHANGE: ("ISM_CHANGE", ">IIBxxx"),
    MSG_NSM_CHANGE: ("NSM_CHANGE", ">IIIBxxx"),
    MSG_REACHABLE_CHANGE: ("REACHABLE_CHANGE", ">HH"),
    MSG_ROUTER_ID_CHANGE: ("ROUTER_ID_CHANGE", ">I"),
}

OSPF_API_OK = 0
OSPF_API_NOSUCHINTERFACE = -1
OSPF_API_NOSUCHAREA = -2
OSPF_API_NOSUCHLSA = -3
OSPF_API_ILLEGALLSATYPE = -4
OSPF_API_OPAQUETYPEINUSE = -5
OSPF_API_OPAQUETYPENOTREGISTERED = -6
OSPF_API_NOTREADY = -7
OSPF_API_NOMEMORY = -8
OSPF_API_ERROR = -9
OSPF_API_UNDEF = -10

msg_errname = {
    OSPF_API_OK: "OSPF_API_OK",
    OSPF_API_NOSUCHINTERFACE: "OSPF_API_NOSUCHINTERFACE",
    OSPF_API_NOSUCHAREA: "OSPF_API_NOSUCHAREA",
    OSPF_API_NOSUCHLSA: "OSPF_API_NOSUCHLSA",
    OSPF_API_ILLEGALLSATYPE: "OSPF_API_ILLEGALLSATYPE",
    OSPF_API_OPAQUETYPEINUSE: "OSPF_API_OPAQUETYPEINUSE",
    OSPF_API_OPAQUETYPENOTREGISTERED: "OSPF_API_OPAQUETYPENOTREGISTERED",
    OSPF_API_NOTREADY: "OSPF_API_NOTREADY",
    OSPF_API_NOMEMORY: "OSPF_API_NOMEMORY",
    OSPF_API_ERROR: "OSPF_API_ERROR",
    OSPF_API_UNDEF: "OSPF_API_UNDEF",
}

# msg_info = {**smsg_info, **amsg_info}
msg_info = {}
msg_info.update(smsg_info)
msg_info.update(amsg_info)
msg_name = {k: v[0] for k, v in msg_info.items()}
msg_fmt = {k: v[1] for k, v in msg_info.items()}
msg_size = {k: struct.calcsize(v) for k, v in msg_fmt.items()}


def api_msgname(mt):
    return msg_name.get(mt, str(mt))


def api_errname(ecode):
    return msg_errname.get(ecode, str(ecode))


# -------------------
# API Semantic Errors
# -------------------


class APIError(Exception):
    pass


class MsgTypeError(Exception):
    pass


class SeqNumError(Exception):
    pass


# ---------
# LSA Types
# ---------

LSA_TYPE_UNKNOWN = 0
LSA_TYPE_ROUTER = 1
LSA_TYPE_NETWORK = 2
LSA_TYPE_SUMMARY = 3
LSA_TYPE_ASBR_SUMMARY = 4
LSA_TYPE_AS_EXTERNAL = 5
LSA_TYPE_GROUP_MEMBER = 6
LSA_TYPE_AS_NSSA = 7
LSA_TYPE_EXTERNAL_ATTRIBUTES = 8
LSA_TYPE_OPAQUE_LINK = 9
LSA_TYPE_OPAQUE_AREA = 10
LSA_TYPE_OPAQUE_AS = 11


def lsa_typename(lsa_type):
    names = {
        LSA_TYPE_ROUTER: "LSA:ROUTER",
        LSA_TYPE_NETWORK: "LSA:NETWORK",
        LSA_TYPE_SUMMARY: "LSA:SUMMARY",
        LSA_TYPE_ASBR_SUMMARY: "LSA:ASBR_SUMMARY",
        LSA_TYPE_AS_EXTERNAL: "LSA:AS_EXTERNAL",
        LSA_TYPE_GROUP_MEMBER: "LSA:GROUP_MEMBER",
        LSA_TYPE_AS_NSSA: "LSA:AS_NSSA",
        LSA_TYPE_EXTERNAL_ATTRIBUTES: "LSA:EXTERNAL_ATTRIBUTES",
        LSA_TYPE_OPAQUE_LINK: "LSA:OPAQUE_LINK",
        LSA_TYPE_OPAQUE_AREA: "LSA:OPAQUE_AREA",
        LSA_TYPE_OPAQUE_AS: "LSA:OPAQUE_AS",
    }
    return names.get(lsa_type, str(lsa_type))


# ------------------------------
# Interface State Machine States
# ------------------------------

ISM_DEPENDUPON = 0
ISM_DOWN = 1
ISM_LOOPBACK = 2
ISM_WAITING = 3
ISM_POINTTOPOINT = 4
ISM_DROTHER = 5
ISM_BACKUP = 6
ISM_DR = 7


def ism_name(state):
    names = {
        ISM_DEPENDUPON: "ISM_DEPENDUPON",
        ISM_DOWN: "ISM_DOWN",
        ISM_LOOPBACK: "ISM_LOOPBACK",
        ISM_WAITING: "ISM_WAITING",
        ISM_POINTTOPOINT: "ISM_POINTTOPOINT",
        ISM_DROTHER: "ISM_DROTHER",
        ISM_BACKUP: "ISM_BACKUP",
        ISM_DR: "ISM_DR",
    }
    return names.get(state, str(state))


# -----------------------------
# Neighbor State Machine States
# -----------------------------

NSM_DEPENDUPON = 0
NSM_DELETED = 1
NSM_DOWN = 2
NSM_ATTEMPT = 3
NSM_INIT = 4
NSM_TWOWAY = 5
NSM_EXSTART = 6
NSM_EXCHANGE = 7
NSM_LOADING = 8
NSM_FULL = 9


def nsm_name(state):
    names = {
        NSM_DEPENDUPON: "NSM_DEPENDUPON",
        NSM_DELETED: "NSM_DELETED",
        NSM_DOWN: "NSM_DOWN",
        NSM_ATTEMPT: "NSM_ATTEMPT",
        NSM_INIT: "NSM_INIT",
        NSM_TWOWAY: "NSM_TWOWAY",
        NSM_EXSTART: "NSM_EXSTART",
        NSM_EXCHANGE: "NSM_EXCHANGE",
        NSM_LOADING: "NSM_LOADING",
        NSM_FULL: "NSM_FULL",
    }
    return names.get(state, str(state))


class WithNothing:
    "An object that does nothing when used with `with` statement."

    async def __aenter__(self):
        return

    async def __aexit__(self, *args, **kwargs):
        return


# --------------
# Client Classes
# --------------


class OspfApiClient:
    def __str__(self):
        return "OspfApiClient({})".format(self.server)

    @staticmethod
    def _get_bound_sockets(port):
        s1 = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        try:
            s1.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            # s1.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            s1.bind(("", port))
            s2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            try:
                s2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                # s2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
                s2.bind(("", port + 1))
                return s1, s2
            except Exception:
                s2.close()
                raise
        except Exception:
            s1.close()
            raise

    def __init__(self, server="localhost", handlers=None):
        """A client connection to OSPF Daemon using the OSPF API

        The client object is not created in a connected state.  To connect to the server
        the `connect` method should be called.  If an error is encountered when sending
        messages to the server an exception will be raised and the connection will be
        closed.  When this happens `connect` may be called again to restore the
        connection.

        Args:
            server: hostname or IP address of server default is "localhost"
            handlers: dict of message handlers, the key is the API message
                type, the value is a function. The functions signature is:
                `handler(msg_type, msg, msg_extra, *params)`, where `msg` is the
                message data after the API header, `*params` will be the
                unpacked message values, and msg_extra are any bytes beyond the
                fixed parameters of the message.
        Raises:
            Will raise exceptions for failures with various `socket` modules
            functions such as `socket.socket`, `socket.setsockopt`, `socket.bind`.
        """
        self._seq = 0
        self._s = None
        self._as = None
        self._ls = None
        self._ar = self._r = self._aw = self._w = None
        self.server = server
        self.handlers = handlers if handlers is not None else dict()
        self.write_lock = Lock()

        # try and get consecutive 2 ports
        PORTSTART = 49152
        PORTEND = 65534
        for port in range(PORTSTART, PORTEND + 2, 2):
            try:
                logging.debug("%s: binding to ports %s, %s", self, port, port + 1)
                self._s, self._ls = self._get_bound_sockets(port)
                break
            except OSError as error:
                if error.errno != errno.EADDRINUSE or port == PORTEND:
                    logging.warning("%s: binding port %s error %s", self, port, error)
                    raise
                logging.debug("%s: ports %s, %s in use.", self, port, port + 1)
        else:
            assert False, "Should not reach this code execution point"

    async def _connect_locked(self):
        logging.debug("%s: connect to OSPF API", self)

        loop = asyncio.get_event_loop()

        self._ls.listen()
        try:
            logging.debug("%s: connecting sync socket to server", self)
            await loop.sock_connect(self._s, (self.server, 2607))

            logging.debug("%s: accepting connect from server", self)
            self._as, _ = await loop.sock_accept(self._ls)
        except Exception:
            await self._close_locked()
            raise

        logging.debug("%s: success", self)
        self._r, self._w = await asyncio.open_connection(sock=self._s)
        self._ar, self._aw = await asyncio.open_connection(sock=self._as)
        self._seq = 1

    async def connect(self):
        async with self.write_lock:
            await self._connect_locked()

    @property
    def closed(self):
        "True if the connection is closed."
        return self._seq == 0

    async def _close_locked(self):
        logging.debug("%s: closing", self)
        if self._s:
            if self._w:
                self._w.close()
                await self._w.wait_closed()
                self._w = None
            else:
                self._s.close()
            self._s = None
            self._r = None
        assert self._w is None
        if self._as:
            self._as.close()
            self._as = None
            self._ar = None
        if self._ls:
            self._ls.close()
            self._ls = None
        self._seq = 0

    async def close(self):
        async with self.write_lock:
            await self._close_locked()

    @staticmethod
    async def _msg_read(r, expseq=-1):
        """Read an OSPF API message from the socket `r`

        Args:
            r: socket to read msg from
            expseq: sequence number to expect or -1 for any.
        Raises:
            Will raise exceptions for failures with various `socket` modules,
            Additionally may raise SeqNumError if unexpected seqnum is received.
        """
        try:
            mh = await r.readexactly(FMT_APIMSGHDR_SIZE)
            v, mt, l, seq = struct.unpack(FMT_APIMSGHDR, mh)
            if v != 1:
                raise Exception("received unexpected OSPF API version {}".format(v))
            if expseq == -1:
                logging.debug("_msg_read: got seq: 0x%x on async read", seq)
            elif seq != expseq:
                raise SeqNumError("rx {} != {}".format(seq, expseq))
            msg = await r.readexactly(l) if l else b""
            return mt, msg
        except asyncio.IncompleteReadError:
            raise EOFError

    async def msg_read(self):
        """Read a message from the async notify channel.

        Raises:
            May raise exceptions for failures with various `socket` modules.
        """
        return await OspfApiClient._msg_read(self._ar, -1)

    async def msg_send(self, mt, mp):
        """Send a message to OSPF API and wait for error code reply.

        Args:
            mt: the messaage type
            mp: the message payload
        Returns:
            error: an OSPF_API_XXX error code, 0 for OK.
        Raises:
            Raises SeqNumError if the synchronous reply is the wrong sequence number;
            MsgTypeError if the synchronous reply is not MSG_REPLY. Also,
            may raise exceptions for failures with various `socket` modules,

            The connection will be closed.
        """
        logging.debug("SEND: %s: sending %s seq 0x%x", self, api_msgname(mt), self._seq)
        mh = struct.pack(FMT_APIMSGHDR, 1, mt, len(mp), self._seq)

        seq = self._seq
        self._seq = seq + 1

        try:
            async with self.write_lock:
                self._w.write(mh + mp)
                await self._w.drain()
                mt, mp = await OspfApiClient._msg_read(self._r, seq)

            if mt != MSG_REPLY:
                raise MsgTypeError(
                    "rx {} != {}".format(api_msgname(mt), api_msgname(MSG_REPLY))
                )

            return struct.unpack(msg_fmt[MSG_REPLY], mp)[0]
        except Exception:
            # We've written data with a sequence number
            await self.close()
            raise

    async def msg_send_raises(self, mt, mp=b"\x00" * 4):
        """Send a message to OSPF API and wait for error code reply.

        Args:
            mt: the messaage type
            mp: the message payload
        Raises:
            APIError if the server replies with an error.

            Also may raise exceptions for failures with various `socket` modules,
            as well as MsgTypeError if the synchronous reply is incorrect.
            The connection will be closed for these non-API error exceptions.
        """
        ecode = await self.msg_send(mt, mp)
        if ecode:
            raise APIError("{} error {}".format(api_msgname(mt), api_errname(ecode)))

    async def handle_async_msg(self, mt, msg):
        if mt not in msg_fmt:
            logging.debug("RECV: %s: unknown async msg type %s", self, mt)
            return

        fmt = msg_fmt[mt]
        sz = msg_size[mt]
        tup = struct.unpack(fmt, msg[:sz])
        extra = msg[sz:]

        if mt not in self.handlers:
            logging.debug(
                "RECV: %s: no handlers for msg type %s", self, api_msgname(mt)
            )
            return

        logging.debug("RECV: %s: calling handler for %s", self, api_msgname(mt))
        await self.handlers[mt](mt, msg, extra, *tup)

    #
    # Client to Server Messaging
    #
    @staticmethod
    def lsa_type_mask(*lsa_types):
        "Return a 16 bit mask for each LSA type passed."
        if not lsa_types:
            return 0xFFFF
        mask = 0
        for t in lsa_types:
            assert 0 < t < 16, "LSA type {} out of range [1, 15]".format(t)
            mask |= 1 << t
        return mask

    @staticmethod
    def lsa_filter(origin, areas, lsa_types):
        """Return an LSA filter.

        Return the filter message bytes based on `origin` the `areas` list and the LSAs
        types in the `lsa_types` list.
        """
        mask = OspfApiClient.lsa_type_mask(*lsa_types)
        narea = len(areas)
        fmt = FMT_LSA_FILTER + ("{}I".format(narea) if narea else "")
        # lsa type mask, origin, number of areas, each area
        return struct.pack(fmt, mask, origin, narea, *areas)

    async def req_lsdb_sync(self):
        "Register for all LSA notifications and request an LSDB synchronoization."
        logging.debug("SEND: %s: request LSDB events", self)
        mp = OspfApiClient.lsa_filter(LSAF_ORIGIN_ANY, [], [])
        await self.msg_send_raises(MSG_REGISTER_EVENT, mp)

        logging.debug("SEND: %s: request LSDB sync", self)
        await self.msg_send_raises(MSG_SYNC_LSDB, mp)

    async def req_reachable_routers(self):
        "Request a dump of all reachable routers."
        logging.debug("SEND: %s: request reachable changes", self)
        await self.msg_send_raises(MSG_SYNC_REACHABLE)

    async def req_ism_states(self):
        "Request a dump of the current ISM states of all interfaces."
        logging.debug("SEND: %s: request ISM changes", self)
        await self.msg_send_raises(MSG_SYNC_ISM)

    async def req_nsm_states(self):
        "Request a dump of the current NSM states of all neighbors."
        logging.debug("SEND: %s: request NSM changes", self)
        await self.msg_send_raises(MSG_SYNC_NSM)

    async def req_router_id_sync(self):
        "Request a dump of the current NSM states of all neighbors."
        logging.debug("SEND: %s: request router ID sync", self)
        await self.msg_send_raises(MSG_SYNC_ROUTER_ID)


class OspfOpaqueClient(OspfApiClient):
    """A client connection to OSPF Daemon for manipulating Opaque LSA data.

    The client object is not created in a connected state.  To connect to the server
    the `connect` method should be called.  If an error is encountered when sending
    messages to the server an exception will be raised and the connection will be
    closed.  When this happens `connect` may be called again to restore the
    connection.

    Args:
        server: hostname or IP address of server default is "localhost"
        wait_ready: if True then wait for OSPF to signal ready, in newer versions
            FRR ospfd is always ready so this overhead can be skipped.
            default is False.

    Raises:
        Will raise exceptions for failures with various `socket` modules
        functions such as `socket.socket`, `socket.setsockopt`, `socket.bind`.
    """

    def __init__(self, server="localhost", wait_ready=False):
        handlers = {
            MSG_LSA_UPDATE_NOTIFY: self._lsa_change_msg,
            MSG_LSA_DELETE_NOTIFY: self._lsa_change_msg,
            MSG_NEW_IF: self._if_msg,
            MSG_DEL_IF: self._if_msg,
            MSG_ISM_CHANGE: self._if_change_msg,
            MSG_NSM_CHANGE: self._nbr_change_msg,
            MSG_REACHABLE_CHANGE: self._reachable_msg,
            MSG_ROUTER_ID_CHANGE: self._router_id_msg,
        }
        if wait_ready:
            handlers[MSG_READY_NOTIFY] = self._ready_msg

        super().__init__(server, handlers)

        self.wait_ready = wait_ready
        self.ready_lock = Lock() if wait_ready else WithNothing()
        self.ready_cond = {
            LSA_TYPE_OPAQUE_LINK: {},
            LSA_TYPE_OPAQUE_AREA: {},
            LSA_TYPE_OPAQUE_AS: {},
        }
        self.router_id = ip(0)
        self.router_id_change_cb = None

        self.lsid_seq_num = {}

        self.lsa_change_cb = None
        self.opaque_change_cb = {}

        self.reachable_routers = set()
        self.reachable_change_cb = None

        self.if_area = {}
        self.ism_states = {}
        self.ism_change_cb = None

        self.nsm_states = {}
        self.nsm_change_cb = None

    async def _register_opaque_data(self, lsa_type, otype):
        async with self.ready_lock:
            cond = self.ready_cond[lsa_type].get(otype)
            assert cond is None, "multiple registers for {} opaque-type {}".format(
                lsa_typename(lsa_type), otype
            )

            logging.debug("register %s opaque-type %s", lsa_typename(lsa_type), otype)

            mt = MSG_REGISTER_OPAQUETYPE
            mp = struct.pack(msg_fmt[mt], lsa_type, otype)
            await self.msg_send_raises(mt, mp)

            # If we are not waiting, mark ready for register check
            if not self.wait_ready:
                self.ready_cond[lsa_type][otype] = True

    async def _handle_msg_loop(self):
        try:
            logging.debug("entering async msg handling loop")
            while True:
                mt, msg = await self.msg_read()
                if mt in amsg_info:
                    await self.handle_async_msg(mt, msg)
                else:
                    mts = api_msgname(mt)
                    logging.warning(
                        "ignoring unexpected msg: %s len: %s", mts, len(msg)
                    )
        except EOFError:
            logging.info("Got EOF from OSPF API server on async notify socket")
            return 2

    @staticmethod
    def _opaque_args(lsa_type, otype, oid, mp):
        lsid = (otype << 24) | oid
        return 0, 0, lsa_type, lsid, 0, 0, 0, FMT_LSA_HEADER_SIZE + len(mp)

    @staticmethod
    def _make_opaque_lsa(lsa_type, otype, oid, mp):
        # /* Make a new LSA from parameters */
        lsa = struct.pack(
            FMT_LSA_HEADER, *OspfOpaqueClient._opaque_args(lsa_type, otype, oid, mp)
        )
        lsa += mp
        return lsa

    async def _ready_msg(self, mt, msg, extra, lsa_type, otype, addr):
        assert self.wait_ready

        if lsa_type == LSA_TYPE_OPAQUE_LINK:
            e = "ifaddr {}".format(ip(addr))
        elif lsa_type == LSA_TYPE_OPAQUE_AREA:
            e = "area {}".format(ip(addr))
        else:
            e = ""
        logging.info(
            "RECV: %s ready notify for %s opaque-type %s%s",
            self,
            lsa_typename(lsa_type),
            otype,
            e,
        )

        # Signal all waiting senders they can send now.
        async with self.ready_lock:
            cond = self.ready_cond[lsa_type].get(otype)
            self.ready_cond[lsa_type][otype] = True

        if cond is True:
            logging.warning(
                "RECV: dup ready received for %s opaque-type %s",
                lsa_typename(lsa_type),
                otype,
            )
        elif cond:
            for evt in cond:
                evt.set()

    async def _if_msg(self, mt, msg, extra, *args):
        if mt == MSG_NEW_IF:
            ifaddr, aid = args
        else:
            assert mt == MSG_DEL_IF
            ifaddr, aid = args[0], 0
        logging.info(
            "RECV: %s ifaddr %s areaid %s", api_msgname(mt), ip(ifaddr), ip(aid)
        )

    async def _if_change_msg(self, mt, msg, extra, ifaddr, aid, state):
        ifaddr = ip(ifaddr)
        aid = ip(aid)

        logging.info(
            "RECV: %s ifaddr %s areaid %s state %s",
            api_msgname(mt),
            ifaddr,
            aid,
            ism_name(state),
        )

        self.if_area[ifaddr] = aid
        self.ism_states[ifaddr] = state

        if self.ism_change_cb:
            self.ism_change_cb(ifaddr, aid, state)

    async def _nbr_change_msg(self, mt, msg, extra, ifaddr, nbraddr, router_id, state):
        ifaddr = ip(ifaddr)
        nbraddr = ip(nbraddr)
        router_id = ip(router_id)

        logging.info(
            "RECV: %s ifaddr %s nbraddr %s router_id %s state %s",
            api_msgname(mt),
            ifaddr,
            nbraddr,
            router_id,
            nsm_name(state),
        )

        if ifaddr not in self.nsm_states:
            self.nsm_states[ifaddr] = {}
        self.nsm_states[ifaddr][(nbraddr, router_id)] = state

        if self.nsm_change_cb:
            self.nsm_change_cb(ifaddr, nbraddr, router_id, state)

    async def _lsa_change_msg(self, mt, msg, extra, ifaddr, aid, is_self, *ls_header):
        (
            lsa_age,  # ls_age,
            _,  # ls_options,
            lsa_type,
            ls_id,
            _,  # ls_adv_router,
            ls_seq,
            _,  # ls_cksum,
            ls_len,
        ) = ls_header

        otype = (ls_id >> 24) & 0xFF

        if mt == MSG_LSA_UPDATE_NOTIFY:
            ts = "update"
        else:
            assert mt == MSG_LSA_DELETE_NOTIFY
            ts = "delete"

        logging.info(
            "RECV: LSA %s msg for LSA %s in area %s seq 0x%x len %s age %s",
            ts,
            ip(ls_id),
            ip(aid),
            ls_seq,
            ls_len,
            lsa_age,
        )
        idx = (lsa_type, otype)

        pre_lsa_size = msg_size[mt] - FMT_LSA_HEADER_SIZE
        lsa = msg[pre_lsa_size:]

        if idx in self.opaque_change_cb:
            self.opaque_change_cb[idx](mt, ifaddr, aid, ls_header, extra, lsa)

        if self.lsa_change_cb:
            self.lsa_change_cb(mt, ifaddr, aid, ls_header, extra, lsa)

    async def _reachable_msg(self, mt, msg, extra, nadd, nremove):
        router_ids = struct.unpack(">{}I".format(nadd + nremove), extra)
        router_ids = [ip(x) for x in router_ids]
        logging.info(
            "RECV: %s added %s removed %s",
            api_msgname(mt),
            router_ids[:nadd],
            router_ids[nadd:],
        )
        self.reachable_routers |= set(router_ids[:nadd])
        self.reachable_routers -= set(router_ids[nadd:])
        logging.info("RECV: %s new set %s", api_msgname(mt), self.reachable_routers)

        if self.reachable_change_cb:
            logging.info("RECV: %s calling callback", api_msgname(mt))
            await self.reachable_change_cb(router_ids[:nadd], router_ids[nadd:])

    async def _router_id_msg(self, mt, msg, extra, router_id):
        router_id = ip(router_id)
        logging.info("RECV: %s router ID %s", api_msgname(mt), router_id)
        old_router_id = self.router_id
        if old_router_id == router_id:
            return

        self.router_id = router_id
        logging.info(
            "RECV: %s new router ID %s older router ID %s",
            api_msgname(mt),
            router_id,
            old_router_id,
        )

        if self.router_id_change_cb:
            logging.info("RECV: %s calling callback", api_msgname(mt))
            await self.router_id_change_cb(router_id, old_router_id)

    async def add_opaque_data(self, addr, lsa_type, otype, oid, data):
        """Add an instance of opaque data.

        Add an instance of opaque data. This call will register for the given
        LSA and opaque type if not already done.

        Args:
            addr: depends on lsa_type, LINK => ifaddr, AREA => area ID, AS => ignored
            lsa_type: LSA_TYPE_OPAQUE_{LINK,AREA,AS}
            otype: (octet) opaque type
            oid: (3 octets) ID of this opaque data
            data: the opaque data
        Raises:
            See `msg_send_raises`
        """
        assert self.ready_cond.get(lsa_type, {}).get(otype) is True, "Not Registered!"

        if lsa_type == LSA_TYPE_OPAQUE_LINK:
            ifaddr, aid = int(addr), 0
        elif lsa_type == LSA_TYPE_OPAQUE_AREA:
            ifaddr, aid = 0, int(addr)
        else:
            assert lsa_type == LSA_TYPE_OPAQUE_AS
            ifaddr, aid = 0, 0

        mt = MSG_ORIGINATE_REQUEST
        msg = struct.pack(
            msg_fmt[mt],
            ifaddr,
            aid,
            *OspfOpaqueClient._opaque_args(lsa_type, otype, oid, data),
        )
        msg += data
        await self.msg_send_raises(mt, msg)

    async def delete_opaque_data(self, addr, lsa_type, otype, oid, flags=0):
        """Delete an instance of opaque data.

        Delete an instance of opaque data. This call will register for the given
        LSA and opaque type if not already done.

        Args:
            addr: depends on lsa_type, LINK => ifaddr, AREA => area ID, AS => ignored
            lsa_type: LSA_TYPE_OPAQUE_{LINK,AREA,AS}
            otype: (octet) opaque type.
            oid: (3 octets) ID of this opaque data
            flags: (octet) optional flags (e.g., OSPF_API_DEL_ZERO_LEN_LSA, defaults to no flags)
        Raises:
            See `msg_send_raises`
        """
        assert self.ready_cond.get(lsa_type, {}).get(otype) is True, "Not Registered!"

        mt = MSG_DELETE_REQUEST
        mp = struct.pack(msg_fmt[mt], int(addr), lsa_type, otype, flags, oid)
        await self.msg_send_raises(mt, mp)

    async def is_registered(self, lsa_type, otype):
        """Determine if an (lsa_type, otype) tuple has been registered with FRR

        This determines if the type has been registered, but not necessarily if it is
        ready, if that is required use the `wait_opaque_ready` metheod.

        Args:
            lsa_type: LSA_TYPE_OPAQUE_{LINK,AREA,AS}
            otype: (octet) opaque type.
        """
        async with self.ready_lock:
            return self.ready_cond.get(lsa_type, {}).get(otype) is not None

    async def register_opaque_data(self, lsa_type, otype, callback=None):
        """Register intent to advertise opaque data.

        The application should wait for the async notificaiton that the server is
        ready to advertise the given opaque data type. The API currently only allows
        a single "owner" of each unique (lsa_type,otype). To wait call `wait_opaque_ready`

        Args:
            lsa_type: LSA_TYPE_OPAQUE_{LINK,AREA,AS}
            otype: (octet) opaque type.
            callback: if given, callback will be called when changes are received for
                LSA of the given (lsa_type, otype). The callbacks signature is:

                `callback(msg_type, ifaddr, area_id, lsa_header, data, lsa)`

                Args:
                    msg_type: MSG_LSA_UPDATE_NOTIFY or MSG_LSA_DELETE_NOTIFY
                    ifaddr: integer identifying an interface (by IP address)
                    area_id: integer identifying an area
                    lsa_header: the LSA header as an unpacked tuple (fmt: ">HBBIILHH")
                    data: the opaque data that follows the LSA header
                    lsa: the octets of the full lsa
        Raises:
            See `msg_send_raises`
        """
        assert not await self.is_registered(
            lsa_type, otype
        ), "Registering registered type"

        if callback:
            self.opaque_change_cb[(lsa_type, otype)] = callback
        elif (lsa_type, otype) in self.opaque_change_cb:
            logging.warning(
                "OSPFCLIENT: register: removing callback for %s opaque-type %s",
                lsa_typename(lsa_type),
                otype,
            )
            del self.opaque_change_cb[(lsa_type, otype)]

        await self._register_opaque_data(lsa_type, otype)

    async def wait_opaque_ready(self, lsa_type, otype):
        async with self.ready_lock:
            cond = self.ready_cond[lsa_type].get(otype)
            if cond is True:
                return

            assert self.wait_ready

            logging.debug(
                "waiting for ready %s opaque-type %s", lsa_typename(lsa_type), otype
            )

            if not cond:
                cond = self.ready_cond[lsa_type][otype] = []

            evt = Event()
            cond.append(evt)

        await evt.wait()
        logging.debug("READY for %s opaque-type %s", lsa_typename(lsa_type), otype)

    async def register_opaque_data_wait(self, lsa_type, otype, callback=None):
        """Register intent to advertise opaque data and wait for ready.

        The API currently only allows a single "owner" of each unique (lsa_type,otype).

        Args:
            lsa_type: LSA_TYPE_OPAQUE_{LINK,AREA,AS}
            otype: (octet) opaque type.
            callback: if given, callback will be called when changes are received for
                LSA of the given (lsa_type, otype). The callbacks signature is:

                `callback(msg_type, ifaddr, area_id, lsa_header, data, lsa)`

                Args:
                    msg_type: MSG_LSA_UPDATE_NOTIFY or MSG_LSA_DELETE_NOTIFY
                    ifaddr: integer identifying an interface (by IP address)
                    area_id: integer identifying an area
                    lsa_header: the LSA header as an unpacked tuple (fmt: ">HBBIILHH")
                    data: the opaque data that follows the LSA header
                    lsa: the octets of the full lsa
        Raises:

            See `msg_send_raises`
        """
        await self.register_opaque_data(lsa_type, otype, callback)
        await self.wait_opaque_ready(lsa_type, otype)

    async def unregister_opaque_data(self, lsa_type, otype):
        """Unregister intent to advertise opaque data.

        This will also cause the server to flush/delete all opaque data of
        the given (lsa_type,otype).

        Args:
            lsa_type: LSA_TYPE_OPAQUE_{LINK,AREA,AS}
            otype: (octet) opaque type.
        Raises:
            See `msg_send_raises`
        """
        assert await self.is_registered(
            lsa_type, otype
        ), "Unregistering unregistered type"

        if (lsa_type, otype) in self.opaque_change_cb:
            del self.opaque_change_cb[(lsa_type, otype)]

        mt = MSG_UNREGISTER_OPAQUETYPE
        mp = struct.pack(msg_fmt[mt], lsa_type, otype)
        await self.msg_send_raises(mt, mp)

    async def monitor_lsa(self, callback=None):
        """Monitor changes to LSAs.

        Args:
            callback: if given, callback will be called when changes are received for
                any LSA. The callback signature is:

                `callback(msg_type, ifaddr, area_id, lsa_header, extra, lsa)`

                Args:
                    msg_type: MSG_LSA_UPDATE_NOTIFY or MSG_LSA_DELETE_NOTIFY
                    ifaddr: integer identifying an interface (by IP address)
                    area_id: integer identifying an area
                    lsa_header: the LSA header as an unpacked tuple (fmt: ">HBBIILHH")
                    extra: the octets that follow the LSA header
                    lsa: the octets of the full lsa
        """
        self.lsa_change_cb = callback
        await self.req_lsdb_sync()

    async def monitor_reachable(self, callback=None):
        """Monitor the set of reachable routers.

        The property `reachable_routers` contains the set() of reachable router IDs
        as integers. This set is updated prior to calling the `callback`

        Args:
            callback: callback will be called when the set of reachable
                routers changes. The callback signature is:

                `callback(added, removed)`

                Args:
                    added: list of integer router IDs being added
                    removed: list of integer router IDs being removed
        """
        self.reachable_change_cb = callback
        await self.req_reachable_routers()

    async def monitor_ism(self, callback=None):
        """Monitor the state of OSPF enabled interfaces.

        Args:
            callback: callback will be called when an interface changes state.
                The callback signature is:

                `callback(ifaddr, area_id, state)`

                Args:
                    ifaddr: integer identifying an interface (by IP address)
                    area_id: integer identifying an area
                    state: ISM_*
        """
        self.ism_change_cb = callback
        await self.req_ism_states()

    async def monitor_nsm(self, callback=None):
        """Monitor the state of OSPF neighbors.

        Args:
            callback: callback will be called when a neighbor changes state.
                The callback signature is:

                `callback(ifaddr, nbr_addr, router_id, state)`

                Args:
                    ifaddr: integer identifying an interface (by IP address)
                    nbr_addr: integer identifying neighbor by IP address
                    router_id: integer identifying neighbor router ID
                    state: NSM_*
        """
        self.nsm_change_cb = callback
        await self.req_nsm_states()

    async def monitor_router_id(self, callback=None):
        """Monitor the OSPF router ID.

        The property `router_id` contains the OSPF urouter ID.
        This value is updated prior to calling the `callback`

        Args:
            callback: callback will be called when the router ID changes.
                The callback signature is:

                `callback(new_router_id, old_router_id)`

                Args:
                    new_router_id: the new router ID
                    old_router_id: the old router ID
        """
        self.router_id_change_cb = callback
        await self.req_router_id_sync()


# ================
# CLI/Script Usage
# ================
def next_action(action_list=None):
    "Get next action from list or STDIN"
    if action_list:
        for action in action_list:
            yield action
    else:
        while True:
            action = input("")
            if not action:
                break
            yield action.strip()


async def async_main(args):
    c = OspfOpaqueClient(args.server)
    await c.connect()

    try:
        # Start handling async messages from server.
        if sys.version_info[1] > 6:
            asyncio.create_task(c._handle_msg_loop())
        else:
            asyncio.get_event_loop().create_task(c._handle_msg_loop())

        await c.req_lsdb_sync()
        await c.req_reachable_routers()
        await c.req_ism_states()
        await c.req_nsm_states()

        for action in next_action(args.actions):
            _s = action.split(",")
            what = _s.pop(False)
            if what.casefold() == "wait":
                stime = int(_s.pop(False))
                logging.info("waiting %s seconds", stime)
                await asyncio.sleep(stime)
                logging.info("wait complete: %s seconds", stime)
                continue
            ltype = int(_s.pop(False))
            if ltype == 11:
                addr = ip(0)
            else:
                aval = _s.pop(False)
                try:
                    addr = ip(int(aval))
                except ValueError:
                    addr = ip(aval)
            oargs = [addr, ltype, int(_s.pop(False)), int(_s.pop(False))]

            if not await c.is_registered(oargs[1], oargs[2]):
                await c.register_opaque_data_wait(oargs[1], oargs[2])

            if what.casefold() == "add":
                try:
                    b = bytes.fromhex(_s.pop(False))
                except IndexError:
                    b = b""
                logging.info("opaque data is %s octets", len(b))
                # Needs to be multiple of 4 in length
                mod = len(b) % 4
                if mod:
                    b += b"\x00" * (4 - mod)
                    logging.info("opaque padding to %s octets", len(b))

                await c.add_opaque_data(*oargs, b)
            else:
                assert what.casefold().startswith("del")
                f = 0
                if len(_s) >= 1:
                    try:
                        f = int(_s.pop(False))
                    except IndexError:
                        f = 0
                await c.delete_opaque_data(*oargs, f)
        if not args.actions or args.exit:
            return 0
    except Exception as error:
        logging.error("async_main: unexpected error: %s", error, exc_info=True)
        return 2

    try:
        logging.info("Sleeping forever")
        while True:
            await asyncio.sleep(120)
    except EOFError:
        logging.info("Got EOF from OSPF API server on async notify socket")
        return 2


def main(*args):
    ap = argparse.ArgumentParser(args)
    ap.add_argument("--logtag", default="CLIENT", help="tag to identify log messages")
    ap.add_argument("--exit", action="store_true", help="Exit after commands")
    ap.add_argument("--server", default="localhost", help="OSPF API server")
    ap.add_argument("-v", "--verbose", action="store_true", help="be verbose")
    ap.add_argument(
        "actions",
        nargs="*",
        help="WAIT,SEC|(ADD|DEL),LSATYPE,[ADDR,],OTYPE,OID,[HEXDATA|DEL_FLAG]",
    )
    args = ap.parse_args()

    level = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(
        level=level,
        format="%(asctime)s %(levelname)s: {}: %(name)s %(message)s".format(
            args.logtag
        ),
    )

    logging.info("ospfclient: starting")

    status = 3
    try:
        if sys.version_info[1] > 6:
            # python >= 3.7
            status = asyncio.run(async_main(args))
        else:
            loop = asyncio.get_event_loop()
            try:
                status = loop.run_until_complete(async_main(args))
            finally:
                loop.close()
    except KeyboardInterrupt:
        logging.info("Exiting, received KeyboardInterrupt in main")
    except Exception as error:
        logging.info("Exiting, unexpected exception %s", error, exc_info=True)
    else:
        logging.info("ospfclient: clean exit")

    return status


if __name__ == "__main__":
    exit_status = main()
    sys.exit(exit_status)



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2022 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

"""
If we overwrite import/export RT list via route-maps or even flush by using
`set extcommunity none`, then we must withdraw old paths from VRFs to avoid
stale paths.
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.common_config import step

pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    tgen.add_router("r1")


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router = tgen.gears["r1"]
    router.cmd_raises("ip link add vrf1 type vrf table 10")
    router.cmd_raises("ip link set up dev vrf1")
    router.cmd_raises("ip link add vrf2 type vrf table 20")
    router.cmd_raises("ip link set up dev vrf2")
    router.load_config(TopoRouter.RD_ZEBRA, os.path.join(CWD, "r1/zebra.conf"))
    router.load_config(TopoRouter.RD_BGP, os.path.join(CWD, "r1/bgpd.conf"))
    router.start()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_vrf_leaking_rt_change_route_maps():
    tgen = get_topogen()

    router = tgen.gears["r1"]

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    def _bgp_check_path():
        output = json.loads(router.vtysh_cmd("show bgp vrf vrf2 ipv4 unicast json"))
        expected = {"routes": {"192.168.100.100/32": [{"nhVrfName": "vrf1"}]}}
        return topotest.json_cmp(output, expected)

    step("Initial converge")
    test_func = functools.partial(_bgp_check_path)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Can't see 192.168.100.100/32 leaked from vrf1 into vrf2."

    step("Overwrite RT list (remove rt 65500:11990 from route-map)")
    router.vtysh_cmd(
        """
        config terminal
        route-map rm permit 10
         set extcommunity rt 65500:10100
        exit
        """
    )

    step("Check if 192.168.100.100/32 was removed from vrf2")
    test_func = functools.partial(_bgp_check_path)
    _, result = topotest.run_and_expect(test_func, not None, count=20, wait=0.5)
    assert result is not None, "192.168.100.100/32 still exists in vrf2 as stale."


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright (c) 2023 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

"""
Test if RIP `passive-interface default` and `no passive-interface IFNAME`
combination works as expected.
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, get_topogen
from lib.common_config import step

pytestmark = [pytest.mark.ripd]


def setup_module(mod):
    topodef = {"s1": ("r1", "r2", "r3")}
    tgen = Topogen(topodef, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_frr_config(os.path.join(CWD, "{}/frr.conf".format(rname)))

    tgen.start_router()


def teardown_module():
    tgen = get_topogen()
    tgen.stop_topology()


def test_rip_passive_interface():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"]
    r2 = tgen.gears["r2"]

    def _show_routes(nh_num):
        output = json.loads(r1.vtysh_cmd("show ip route 10.10.10.1/32 json"))
        expected = {
            "10.10.10.1/32": [
                {
                    "internalNextHopNum": nh_num,
                    "internalNextHopActiveNum": nh_num,
                }
            ]
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_show_routes, 2)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assert result is None, "Got 10.10.10.1/32, but the next-hop count is not 2"

    step("Configure `passive-interface default` on r2")
    r2.vtysh_cmd(
        """
    configure terminal
        router rip
            passive-interface default
    """
    )

    test_func = functools.partial(_show_routes, 1)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assert result is None, "Got 10.10.10.1/32, but the next-hop count is not 1"

    step("Configure `no passive-interface r2-eth0` on r2 towards r1")
    r2.vtysh_cmd(
        """
    configure terminal
        router rip
            no passive-interface r2-eth0
    """
    )

    test_func = functools.partial(_show_routes, 2)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assert result is None, "Got 10.10.10.1/32, but the next-hop count is not 2"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python

#
# Copyright 2022 6WIND S.A.
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted, provided
# that the above copyright notice and this permission notice appear
# in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NETDEF DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NETDEF BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
# OF THIS SOFTWARE.
#

"""
Check if fec nexthop resolution works correctly.
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.common_config import step

pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    """
    r1 ---- r2 ---- r3 ---- r4 ----- r5 ---- r6 ---- r7
     <--- ospf ----> <---- isis -----> <--- ospf ---->
    """
    for routern in range(1, 8):
        tgen.add_router("r{}".format(routern))

    switch1 = tgen.add_switch("s1")
    switch1.add_link(tgen.gears["r1"])
    switch1.add_link(tgen.gears["r2"])

    switch2 = tgen.add_switch("s2")
    switch2.add_link(tgen.gears["r2"])
    switch2.add_link(tgen.gears["r3"])

    switch3 = tgen.add_switch("s3")
    switch3.add_link(tgen.gears["r3"])
    switch3.add_link(tgen.gears["r4"])

    switch4 = tgen.add_switch("s4")
    switch4.add_link(tgen.gears["r4"])
    switch4.add_link(tgen.gears["r5"])

    switch5 = tgen.add_switch("s5")
    switch5.add_link(tgen.gears["r5"])
    switch5.add_link(tgen.gears["r6"])

    switch6 = tgen.add_switch("s6")
    switch6.add_link(tgen.gears["r6"])
    switch6.add_link(tgen.gears["r7"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    def _enable_mpls_misc(router):
        router.run("modprobe mpls_router")
        router.run("echo 100000 > /proc/sys/net/mpls/platform_labels")
        router.run("echo 1 > /proc/sys/net/mpls/conf/lo/input")

    router = tgen.gears["r1"]
    _enable_mpls_misc(router)

    router = tgen.gears["r2"]
    _enable_mpls_misc(router)

    router = tgen.gears["r3"]
    _enable_mpls_misc(router)

    router = tgen.gears["r4"]
    _enable_mpls_misc(router)

    router = tgen.gears["r5"]
    _enable_mpls_misc(router)

    router = tgen.gears["r6"]
    _enable_mpls_misc(router)

    router = tgen.gears["r7"]
    _enable_mpls_misc(router)

    for i, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        if rname in ("r1", "r3", "r5", "r7"):
            router.load_config(
                TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
            )
        if rname in ("r3", "r4", "r5"):
            router.load_config(
                TopoRouter.RD_ISIS, os.path.join(CWD, "{}/isisd.conf".format(rname))
            )
        if rname in ("r1", "r2", "r3", "r5", "r6", "r7"):
            router.load_config(
                TopoRouter.RD_OSPF, os.path.join(CWD, "{}/ospfd.conf".format(rname))
            )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


# There are some startup issued when initialising OSPF
# To avoid those issues, load the ospf configuration after zebra started
def test_zebra_fec_nexthop_resolution_finalise_ospf_config():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    topotest.sleep(2)

    tgen.net["r1"].cmd("vtysh -f {}/r1/ospfd.conf.after".format(CWD))
    tgen.net["r2"].cmd("vtysh -f {}/r2/ospfd.conf.after".format(CWD))
    tgen.net["r3"].cmd("vtysh -f {}/r3/ospfd.conf.after".format(CWD))
    tgen.net["r5"].cmd("vtysh -f {}/r5/ospfd.conf.after".format(CWD))
    tgen.net["r6"].cmd("vtysh -f {}/r6/ospfd.conf.after".format(CWD))
    tgen.net["r7"].cmd("vtysh -f {}/r7/ospfd.conf.after".format(CWD))


def test_zebra_fec_nexthop_resolution_bgp():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    def _check_bgp_session():
        r1 = tgen.gears["r1"]

        output = json.loads(r1.vtysh_cmd("show bgp summary json"))

        if output["ipv4Unicast"]["peers"]["192.0.2.7"]["state"] == "Established":
            return None
        return False

    tgen.gears["r3"].vtysh_cmd("config \n no mpls fec nexthop-resolution \n end")
    tgen.gears["r3"].vtysh_cmd("config \n mpls fec nexthop-resolution \n end")
    tgen.gears["r5"].vtysh_cmd("config \n no mpls fec nexthop-resolution \n end")
    tgen.gears["r5"].vtysh_cmd("config \n mpls fec nexthop-resolution \n end")

    test_func1 = functools.partial(_check_bgp_session)
    _, result1 = topotest.run_and_expect(test_func1, None, count=60, wait=0.5)
    assert result1 is None, "Failed to verify the fec_nexthop_resolution: bgp session"


def test_zebra_fec_nexthop_resolution_ping():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    def _check_ping_launch():
        r1 = tgen.gears["r1"]

        ping_launch = "ping 192.0.2.7 -I 192.0.2.1 -c 1"
        selected_lines = r1.run(ping_launch).splitlines()[-2:-1]
        rtx_stats = "".join(selected_lines[0].split(",")[0:3])
        current = topotest.normalize_text(rtx_stats)

        expected_stats = "1 packets transmitted 1 received 0% packet loss"
        expected = topotest.normalize_text(expected_stats)

        if current == expected:
            return None

        return False

    test_func2 = functools.partial(_check_ping_launch)
    _, result2 = topotest.run_and_expect(test_func2, None, count=60, wait=1)
    assert result2 is None, "Failed to verify the fec_nexthop_resolution: ping"


def test_zebra_fec_nexthop_resolution_table():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    def _zebra_check_mpls_table():
        r3 = tgen.gears["r3"]
        inLabel = 0
        outLabels = 0

        """
        Retrieve inLabel from MPLS FEC table
        """
        mpls_fec = r3.vtysh_cmd("show mpls fec 192.0.2.7/32")
        lines = mpls_fec.split("\n")
        for line in lines:
            if "Label" in line:
                inLabel = line.split(": ", 1)[1]

        """
        Retrieve outLabel from BGP
        """
        output = json.loads(r3.vtysh_cmd("show ip route 192.0.2.7/32 json"))

        outLabels = output["192.0.2.7/32"][0]["nexthops"][1]["labels"]

        if (inLabel == 0) or (outLabels == 0):
            return True

        """
        Compare expected data with real data
        """
        output = json.loads(r3.vtysh_cmd("show mpls table " + str(inLabel) + " json"))

        expected = {
            "inLabel": int(inLabel),
            "installed": True,
            "nexthops": [
                {
                    "type": "BGP",
                    "outLabel": outLabels[0],
                    "outLabelStack": outLabels,
                    "distance": 20,
                    "installed": True,
                    "nexthop": "192.168.3.4",
                }
            ],
        }
        return topotest.json_cmp(output, expected)

    test_func3 = functools.partial(_zebra_check_mpls_table)
    _, result3 = topotest.run_and_expect(test_func3, None, count=60, wait=0.5)
    assert result3 is None, "Failed to verify the fec_nexthop_resolution: mpls table"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_bfd_vrf_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2018 by
# Network Device Education Foundation, Inc. ("NetDEF")
# Copyright (c) 2019 by 6WIND
#

"""
test_bfd_vrf_topo1.py: Test the FRR BFD daemon.
"""

import os
import sys
import json
from functools import partial
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.

pytestmark = [pytest.mark.bfdd, pytest.mark.bgpd]


def build_topo(tgen):
    # Create 4 routers
    for routern in range(1, 5):
        tgen.add_router("r{}".format(routern))

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])

    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r4"])


def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    # check for zebra capability
    for _, router in router_list.items():
        if router.check_capability(TopoRouter.RD_ZEBRA, "--vrfwnetns") == False:
            return pytest.skip(
                "Skipping BFD Topo1 VRF NETNS feature. VRF NETNS backend not available on FRR"
            )

    if os.system("ip netns list") != 0:
        return pytest.skip(
            "Skipping BFD Topo1 VRF NETNS Test. NETNS not available on System"
        )

    logger.info("Testing with VRF Namespace support")

    for rname, router in router_list.items():
        # create VRF rx-bfd-cust1 and link rx-eth0 to rx-bfd-cust1
        ns = "{}-bfd-cust1".format(rname)
        router.net.add_netns(ns)
        router.net.set_intf_netns(rname + "-eth0", ns, up=True)
        if rname == "r2":
            router.net.set_intf_netns(rname + "-eth1", ns, up=True)
            router.net.set_intf_netns(rname + "-eth2", ns, up=True)

    for rname, router in router_list.items():
        router.use_netns_vrf()
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BFD, os.path.join(CWD, "{}/bfdd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    # Initialize all routers.
    tgen.start_router()


def teardown_module(_mod):
    "Teardown the pytest environment"
    tgen = get_topogen()

    # Move interfaces out of vrf namespace and delete the namespace
    router_list = tgen.routers()
    for rname, router in router_list.items():
        if rname == "r2":
            router.net.reset_intf_netns(rname + "-eth2")
            router.net.reset_intf_netns(rname + "-eth1")
        router.net.reset_intf_netns(rname + "-eth0")
        router.net.delete_netns("{}-bfd-cust1".format(rname))
    tgen.stop_topology()


def test_bfd_connection():
    "Assert that the BFD peers can find themselves."
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("waiting for bfd peers to go up")
    for router in tgen.routers().values():
        json_file = "{}/{}/peers.json".format(CWD, router.name)
        expected = json.loads(open(json_file).read())

        test_func = partial(
            topotest.router_json_cmp, router, "show bfd peers json", expected
        )
        _, result = topotest.run_and_expect(test_func, None, count=16, wait=1)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert result is None, assertmsg


def test_bgp_convergence():
    "Assert that BGP is converging."
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("waiting for bgp peers to go up")

    for router in tgen.routers().values():
        ref_file = "{}/{}/bgp_summary.json".format(CWD, router.name)
        expected = json.loads(open(ref_file).read())
        test_func = partial(
            topotest.router_json_cmp,
            router,
            "show ip bgp vrf {}-bfd-cust1 summary json".format(router.name),
            expected,
        )
        _, res = topotest.run_and_expect(test_func, None, count=125, wait=1.0)
        assertmsg = "{}: bgp did not converge".format(router.name)
        assert res is None, assertmsg


def test_bgp_fast_convergence():
    "Assert that BGP is converging before setting a link down."
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("waiting for bgp peers converge")

    for router in tgen.routers().values():
        ref_file = "{}/{}/bgp_prefixes.json".format(CWD, router.name)
        expected = json.loads(open(ref_file).read())
        test_func = partial(
            topotest.router_json_cmp,
            router,
            "show ip bgp vrf {}-bfd-cust1 json".format(router.name),
            expected,
        )
        _, res = topotest.run_and_expect(test_func, None, count=40, wait=1)
        assertmsg = "{}: bgp did not converge".format(router.name)
        assert res is None, assertmsg


def test_bfd_fast_convergence():
    """
    Assert that BFD notices the link down after simulating network
    failure.
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Disable r2-eth0 link
    router2 = tgen.gears["r2"]
    topotest.interface_set_status(
        router2, "r2-eth0", ifaceaction=False, vrf_name="r2-bfd-cust1"
    )

    # Wait the minimum time we can before checking that BGP/BFD
    # converged.
    logger.info("waiting for BFD converge")

    # Check that BGP converged quickly.
    for router in tgen.routers().values():
        json_file = "{}/{}/peers.json".format(CWD, router.name)
        expected = json.loads(open(json_file).read())

        # Load the same file as previous test, but expect R1 to be down.
        if router.name == "r1":
            for peer in expected:
                if peer["peer"] == "192.168.0.2":
                    peer["status"] = "down"
        else:
            for peer in expected:
                if peer["peer"] == "192.168.0.1":
                    peer["status"] = "down"

        test_func = partial(
            topotest.router_json_cmp, router, "show bfd peers json", expected
        )
        _, res = topotest.run_and_expect(test_func, None, count=40, wait=1)
        assertmsg = '"{}" JSON output mismatches'.format(router.name)
        assert res is None, assertmsg


def test_bgp_fast_reconvergence():
    "Assert that BGP is converging after setting a link down."
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("waiting for BGP re convergence")

    # Check that BGP converged quickly.
    for router in tgen.routers().values():
        ref_file = "{}/{}/bgp_prefixes.json".format(CWD, router.name)
        expected = json.loads(open(ref_file).read())

        # Load the same file as previous test, but set networks to None
        # to test absence.
        if router.name == "r1":
            expected["routes"]["10.254.254.2/32"] = None
            expected["routes"]["10.254.254.3/32"] = None
            expected["routes"]["10.254.254.4/32"] = None
        else:
            expected["routes"]["10.254.254.1/32"] = None

        test_func = partial(
            topotest.router_json_cmp,
            router,
            "show ip bgp vrf {}-bfd-cust1 json".format(router.name),
            expected,
        )
        _, res = topotest.run_and_expect(test_func, None, count=16, wait=1)
        assertmsg = "{}: bgp did not converge".format(router.name)
        assert res is None, assertmsg


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_msdp_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2021 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
test_msdp_topo1.py: Test the FRR PIM MSDP peer.
"""

import os
import sys
import json
from functools import partial
import re
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest

# Required to instantiate the topology builder class.
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

from lib.pim import McastTesterHelper

pytestmark = [pytest.mark.bgpd, pytest.mark.pimd]

app_helper = McastTesterHelper()


def build_topo(tgen):
    "Build function"

    # Create 4 routers
    for routern in range(1, 5):
        tgen.add_router("r{}".format(routern))

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r3"])

    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r4"])

    switch = tgen.add_switch("s4")
    # switch.add_link(tgen.gears["r3"])
    switch.add_link(tgen.gears["r4"])

    switch = tgen.add_switch("s5")
    switch.add_link(tgen.gears["r4"])

    # Create a host connected and direct at r4:
    tgen.add_host("h1", "192.168.4.100/24", "via 192.168.4.1")
    tgen.add_host("h3", "192.168.4.120/24", "via 192.168.4.1")
    switch.add_link(tgen.gears["h1"])
    switch.add_link(tgen.gears["h3"])

    # Create a host connected and direct at r1:
    switch = tgen.add_switch("s6")
    tgen.add_host("h2", "192.168.10.100/24", "via 192.168.10.1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["h2"])


def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    for rname, router in router_list.items():
        daemon_file = "{}/{}/zebra.conf".format(CWD, rname)
        if os.path.isfile(daemon_file):
            router.load_config(TopoRouter.RD_ZEBRA, daemon_file)

        daemon_file = "{}/{}/bgpd.conf".format(CWD, rname)
        if os.path.isfile(daemon_file):
            router.load_config(TopoRouter.RD_BGP, daemon_file)

        daemon_file = "{}/{}/pimd.conf".format(CWD, rname)
        if os.path.isfile(daemon_file):
            router.load_config(TopoRouter.RD_PIM, daemon_file)

    # Initialize all routers.
    tgen.start_router()

    app_helper.init(tgen)


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()
    app_helper.cleanup()
    tgen.stop_topology()


def test_bgp_convergence():
    "Wait for BGP protocol convergence"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("waiting for protocols to converge")

    def expect_loopback_route(router, iptype, route, proto):
        "Wait until route is present on RIB for protocol."
        logger.info("waiting route {} in {}".format(route, router))
        test_func = partial(
            topotest.router_json_cmp,
            tgen.gears[router],
            "show {} route json".format(iptype),
            {route: [{"protocol": proto}]},
        )
        _, result = topotest.run_and_expect(test_func, None, count=130, wait=1)
        assertmsg = '"{}" convergence failure'.format(router)
        assert result is None, assertmsg

    # Wait for R1
    expect_loopback_route("r1", "ip", "10.254.254.2/32", "bgp")
    expect_loopback_route("r1", "ip", "10.254.254.3/32", "bgp")
    expect_loopback_route("r1", "ip", "10.254.254.4/32", "bgp")

    # Wait for R2
    expect_loopback_route("r2", "ip", "10.254.254.1/32", "bgp")
    expect_loopback_route("r2", "ip", "10.254.254.3/32", "bgp")
    expect_loopback_route("r2", "ip", "10.254.254.4/32", "bgp")

    # Wait for R3
    expect_loopback_route("r3", "ip", "10.254.254.1/32", "bgp")
    expect_loopback_route("r3", "ip", "10.254.254.2/32", "bgp")
    expect_loopback_route("r3", "ip", "10.254.254.4/32", "bgp")

    # Wait for R4
    expect_loopback_route("r4", "ip", "10.254.254.1/32", "bgp")
    expect_loopback_route("r4", "ip", "10.254.254.2/32", "bgp")
    expect_loopback_route("r4", "ip", "10.254.254.3/32", "bgp")


def _test_mroute_install():
    "Test that multicast routes propagated and installed"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    #
    # Test R1 mroute
    #
    expect_1 = {
        "229.1.2.3": {
            "192.168.10.100": {
                "iif": "r1-eth2",
                "flags": "SFT",
                "oil": {
                    "r1-eth0": {"source": "192.168.10.100", "group": "229.1.2.3"},
                    "r1-eth1": None,
                },
            }
        }
    }
    # Create a deep copy of `expect_1`.
    expect_2 = json.loads(json.dumps(expect_1))
    # The route will be either via R2 or R3.
    expect_2["229.1.2.3"]["192.168.10.100"]["oil"]["r1-eth0"] = None
    expect_2["229.1.2.3"]["192.168.10.100"]["oil"]["r1-eth1"] = {
        "source": "192.168.10.100",
        "group": "229.1.2.3",
    }

    def test_r1_mroute():
        "Test r1 multicast routing table function"
        out = tgen.gears["r1"].vtysh_cmd("show ip mroute json", isjson=True)
        if topotest.json_cmp(out, expect_1) is None:
            return None
        return topotest.json_cmp(out, expect_2)

    logger.info("Waiting for R1 multicast routes")
    _, val = topotest.run_and_expect(test_r1_mroute, None, count=55, wait=2)
    assert val is None, "multicast route convergence failure"

    #
    # Test routers 2 and 3.
    #
    # NOTE: only one of the paths will get the multicast route.
    #
    expect_r2 = {
        "229.1.2.3": {
            "192.168.10.100": {
                "iif": "r2-eth0",
                "flags": "S",
                "oil": {
                    "r2-eth1": {
                        "source": "192.168.10.100",
                        "group": "229.1.2.3",
                    }
                },
            }
        }
    }
    expect_r3 = {
        "229.1.2.3": {
            "192.168.10.100": {
                "iif": "r3-eth0",
                "flags": "S",
                "oil": {
                    "r3-eth1": {
                        "source": "192.168.10.100",
                        "group": "229.1.2.3",
                    }
                },
            }
        }
    }

    def test_r2_r3_mroute():
        "Test r2/r3 multicast routing table function"
        r2_out = tgen.gears["r2"].vtysh_cmd("show ip mroute json", isjson=True)
        r3_out = tgen.gears["r3"].vtysh_cmd("show ip mroute json", isjson=True)

        if topotest.json_cmp(r2_out, expect_r2) is not None:
            return topotest.json_cmp(r3_out, expect_r3)

        return topotest.json_cmp(r2_out, expect_r2)

    logger.info("Waiting for R2 and R3 multicast routes")
    _, val = topotest.run_and_expect(test_r2_r3_mroute, None, count=55, wait=2)
    assert val is None, "multicast route convergence failure"

    #
    # Test router 4
    #
    expect_4 = {
        "229.1.2.3": {
            "*": {
                "iif": "lo",
                "flags": "SC",
                "oil": {
                    "pimreg": {
                        "source": "*",
                        "group": "229.1.2.3",
                        "inboundInterface": "lo",
                        "outboundInterface": "pimreg",
                    },
                    "r4-eth2": {
                        "source": "*",
                        "group": "229.1.2.3",
                        "inboundInterface": "lo",
                        "outboundInterface": "r4-eth2",
                    },
                },
            },
            "192.168.10.100": {
                "iif": "r4-eth0",
                "flags": "ST",
                "oil": {
                    "r4-eth2": {
                        "source": "192.168.10.100",
                        "group": "229.1.2.3",
                        "inboundInterface": "r4-eth0",
                        "outboundInterface": "r4-eth2",
                    }
                },
            },
        }
    }

    test_func = partial(
        topotest.router_json_cmp,
        tgen.gears["r4"],
        "show ip mroute json",
        expect_4,
    )
    logger.info("Waiting for R4 multicast routes")
    _, val = topotest.run_and_expect(test_func, None, count=55, wait=2)
    assert val is None, "multicast route convergence failure"


def test_mroute_install():
    tgen = get_topogen()
    # pytest.skip("FOO")
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Starting helper1")
    mcastaddr = "229.1.2.3"
    app_helper.run("h1", [mcastaddr, "h1-eth0"])

    logger.info("Starting helper2")
    app_helper.run("h2", ["--send=0.7", mcastaddr, "h2-eth0"])

    _test_mroute_install()


def test_msdp():
    """
    Test MSDP convergence.

    MSDP non meshed groups must propagate the whole SA database (not just
    their own) to all peers because not all peers talk with each other.

    This setup leads to a potential loop that can be prevented by checking
    the route's first AS in AS path: it must match the remote eBGP AS number.
    """
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1_expect = {
        "192.168.0.2": {
            "peer": "192.168.0.2",
            "local": "192.168.0.1",
            "state": "established",
        },
        "192.168.1.2": {
            "peer": "192.168.1.2",
            "local": "192.168.1.1",
            "state": "established",
        },
    }
    r1_sa_expect = {
        "229.1.2.3": {
            "192.168.10.100": {
                "source": "192.168.10.100",
                "group": "229.1.2.3",
                "rp": "-",
                "local": "yes",
                "sptSetup": "-",
            }
        }
    }
    r2_expect = {
        "192.168.0.1": {
            "peer": "192.168.0.1",
            "local": "192.168.0.2",
            "state": "established",
        },
        "192.168.2.2": {
            "peer": "192.168.2.2",
            "local": "192.168.2.1",
            "state": "established",
        },
    }
    # Only R2 or R3 will get this SA.
    r2_r3_sa_expect = {
        "229.1.2.3": {
            "192.168.10.100": {
                "source": "192.168.10.100",
                "group": "229.1.2.3",
                "rp": "10.254.254.1",
                "local": "no",
                "sptSetup": "no",
            }
        }
    }
    r3_expect = {
        "192.168.1.1": {
            "peer": "192.168.1.1",
            "local": "192.168.1.2",
            "state": "established",
        },
        # "192.169.3.2": {
        #    "peer": "192.168.3.2",
        #    "local": "192.168.3.1",
        #    "state": "established"
        # }
    }
    r4_expect = {
        "192.168.2.1": {
            "peer": "192.168.2.1",
            "local": "192.168.2.2",
            "state": "established",
        },
        # "192.168.3.1": {
        #    "peer": "192.168.3.1",
        #    "local": "192.168.3.2",
        #    "state": "established"
        # }
    }
    r4_sa_expect = {
        "229.1.2.3": {
            "192.168.10.100": {
                "source": "192.168.10.100",
                "group": "229.1.2.3",
                "rp": "10.254.254.1",
                "local": "no",
                "sptSetup": "yes",
            }
        }
    }

    for router in [
        ("r1", r1_expect, r1_sa_expect),
        ("r2", r2_expect, r2_r3_sa_expect),
        ("r3", r3_expect, r2_r3_sa_expect),
        ("r4", r4_expect, r4_sa_expect),
    ]:
        test_func = partial(
            topotest.router_json_cmp,
            tgen.gears[router[0]],
            "show ip msdp peer json",
            router[1],
        )
        logger.info("Waiting for {} msdp peer data".format(router[0]))
        _, val = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assert val is None, "multicast route convergence failure"

        test_func = partial(
            topotest.router_json_cmp,
            tgen.gears[router[0]],
            "show ip msdp sa json",
            router[2],
        )
        logger.info("Waiting for {} msdp SA data".format(router[0]))
        _, val = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assert val is None, "multicast route convergence failure"


def test_msdp_sa_filter():
    "Start a number of multicast streams and check if filtering works"

    tgen = get_topogen()

    # Flow from r1 -> r4
    for multicast_address in ["229.2.1.1", "229.2.1.2", "229.2.2.1"]:
        app_helper.run("h1", [multicast_address, "h1-eth0"])
        app_helper.run("h2", ["--send=0.7", multicast_address, "h2-eth0"])

    # Flow from r4 -> r1
    for multicast_address in ["229.3.1.1", "229.3.1.2", "229.3.2.1"]:
        app_helper.run("h1", ["--send=0.7", multicast_address, "h1-eth0"])
        app_helper.run("h2", [multicast_address, "h2-eth0"])

    # Flow from r4 -> r1 but with more sources
    for multicast_address in ["229.10.1.1", "229.11.1.1"]:
        app_helper.run("h1", ["--send=0.7", multicast_address, "h1-eth0"])
        app_helper.run("h2", [multicast_address, "h2-eth0"])
        app_helper.run("h3", ["--send=0.7", multicast_address, "h3-eth0"])

    # Test that we don't learn any filtered multicast streams.
    r4_sa_expected = {
        "229.2.1.1": None,
        "229.2.1.2": None,
        "229.2.2.1": {
            "192.168.10.100": {
                "local": "no",
                "sptSetup": "yes",
            }
        },
    }
    test_func = partial(
        topotest.router_json_cmp,
        tgen.gears["r4"],
        "show ip msdp sa json",
        r4_sa_expected,
    )
    logger.info("Waiting for r4 MDSP SA data")
    _, val = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert val is None, "multicast route convergence failure"

    # Test that we don't send any filtered multicast streams.
    r1_sa_expected = {
        "229.3.1.1": None,
        "229.3.1.2": None,
        "229.3.2.1": {
            "192.168.4.100": {
                "local": "no",
                "sptSetup": "yes",
            }
        },
        "229.10.1.1": {
            "192.168.4.100": None,
            "192.168.4.120": {
                "local": "no",
                "sptSetup": "yes",
            },
        },
        "229.11.1.1": {
            "192.168.4.100": {
                "local": "no",
                "sptSetup": "yes",
            },
            "192.168.4.120": {
                "local": "no",
                "sptSetup": "yes",
            },
        },
    }
    test_func = partial(
        topotest.router_json_cmp,
        tgen.gears["r1"],
        "show ip msdp sa json",
        r1_sa_expected,
    )
    logger.info("Waiting for r1 MDSP SA data")
    _, val = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert val is None, "multicast route convergence failure"


def test_msdp_sa_limit():
    "Test MSDP SA limiting."

    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    tgen.gears["r4"].vtysh_cmd(
        """
    configure terminal
    router pim
     msdp log sa-events
     msdp peer 192.168.2.1 sa-limit 4
     msdp peer 192.168.3.1 sa-limit 4
    """
    )

    # Flow from r1 -> r4
    for multicast_address in [
        "229.1.2.10",
        "229.1.2.11",
        "229.1.2.12",
        "229.1.2.13",
        "229.1.2.14",
    ]:
        app_helper.run("h1", [multicast_address, "h1-eth0"])
        app_helper.run("h2", ["--send=0.7", multicast_address, "h2-eth0"])

    def test_sa_limit_log():
        r4_log = tgen.gears["r4"].net.getLog("log", "pimd")
        return re.search(r"MSDP peer .+ reject SA (.+, .+): SA limit \d+ of 4", r4_log)

    _, val = topotest.run_and_expect(test_sa_limit_log, None, count=30, wait=1)
    assert val is None, "SA limit check failed"


def test_msdp_log_events():
    "Test that the enabled logs are working as expected."

    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1_log = tgen.gears["r1"].net.getLog("log", "pimd")

    # Look up for informational messages that should have been enabled.
    match = re.search("MSDP peer 192.168.1.2 state changed to established", r1_log)
    assert match is not None

    match = re.search(r"MSDP SA \(192.168.10.100\,229.1.2.3\) created", r1_log)
    assert match is not None


def test_msdp_shutdown():
    "Shutdown MSDP sessions between r1, r2, r3, then check the state."

    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    tgen.gears["r1"].vtysh_cmd(
        """
    configure terminal
    router pim
     msdp shutdown
    """
    )

    r1_expect = {
        "192.168.0.2": {
            "state": "inactive",
        },
        "192.168.1.2": {
            "state": "inactive",
        },
    }
    r2_expect = {
        "192.168.0.1": {
            "state": "listen",
        }
    }
    r3_expect = {
        "192.168.1.1": {
            "state": "listen",
        }
    }
    for router in [("r1", r1_expect), ("r2", r2_expect), ("r3", r3_expect)]:
        test_func = partial(
            topotest.router_json_cmp,
            tgen.gears[router[0]],
            "show ip msdp peer json",
            router[1],
        )
        logger.info("Waiting for {} msdp peer data".format(router[0]))
        _, val = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assert val is None, "multicast route convergence failure"


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_ospf_prefix_suppression.py
#
# Copyright (c) 2023 LabN Consulting
# Acee Lindem
#

import os
import sys
from functools import partial
import pytest

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, get_topogen
from lib.topolog import logger

from lib.common_config import (
    step,
)


"""
test_ospf_metric_propagation.py: Test OSPF/BGP metric propagation
"""

TOPOLOGY = """


            +-----+             +-----+              +-----+
       eth4 |     |   eth0      |     | eth4    eth0 |     |
      ------+     +-------------+     +--------------+     |
10.1.7.0/24 |     | 10.1.1.0/24 |     | 10.1.5.0/24  |     |
            |     |             |     |.2          .3|     |
            |     |   eth1      |     |              |     |
            |     +-------------+     |              |     |
            | R1  | 10.1.2.0/24 |  R2 |              |  R3 |
            |     |             |     |              |     |
            |     |   eth2      |     |              |     |
            |     +-------------+     |              |     |
            |     | 10.1.3.0/24 |     |              |     |
            |     |             |     |              |     |
            |     |   eth3      |     | eth5    eth1 |     |
            |     +-------------+     +--------------+     |
            |     | 10.1.4.0/24 |     | 10.1.6.0/24  |     |
         .1 +-----+.1         .2+-----+.2          .3+-----+

"""

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# Required to instantiate the topology builder class.

pytestmark = [pytest.mark.ospfd, pytest.mark.bgpd]


def build_topo(tgen):
    "Build function"

    # Create 3 routers
    tgen.add_router("r1")
    tgen.add_router("r2")
    tgen.add_router("r3")

    # Interconect router 1, 2 (0)
    switch = tgen.add_switch("s1-1-2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    # Interconect router 1, 2 (1)
    switch = tgen.add_switch("s2-1-2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    # Interconect router 1, 2 (2)
    switch = tgen.add_switch("s3-1-2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    # Interconect router 1, 2 (3)
    switch = tgen.add_switch("s4-1-2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    # Interconect router 2, 3 (0)
    switch = tgen.add_switch("s5-2-3")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])

    # Interconect router 2, 3 (1)
    switch = tgen.add_switch("s6-2-3")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])

    # Add standalone network to router 1
    switch = tgen.add_switch("s7-1")
    switch.add_link(tgen.gears["r1"])


def setup_module(mod):
    logger.info("OSPF Prefix Suppression:\n {}".format(TOPOLOGY))

    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    # Starting Routers
    router_list = tgen.routers()

    for rname, router in router_list.items():
        logger.info("Loading router %s" % rname)
        router.load_frr_config(os.path.join(CWD, "{}/frr.conf".format(rname)))

    # Initialize all routers.
    tgen.start_router()


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()
    tgen.stop_topology()


def test_all_routes_advertised():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("Skipped because of router(s) failure")

    # Verify OSPF routes are installed
    r3 = tgen.gears["r3"]
    input_dict = {
        "10.1.1.0/24": [
            {
                "prefix": "10.1.1.0/24",
                "prefixLen": 24,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.1.0/24 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.1.0/24 not installed on router r3"
    assert result is None, assertmsg

    input_dict = {
        "10.1.2.0/24": [
            {
                "prefix": "10.1.2.0/24",
                "prefixLen": 24,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.2.0/24 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.2.0/24 not installed on router r3"
    assert result is None, assertmsg

    input_dict = {
        "10.1.3.0/24": [
            {
                "prefix": "10.1.3.0/24",
                "prefixLen": 24,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.3.0/24 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.3.0/24 not installed on router r3"
    assert result is None, assertmsg

    input_dict = {
        "10.1.4.1/32": [
            {
                "prefix": "10.1.4.1/32",
                "prefixLen": 32,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.4.1/32 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.4.1/32 not installed on router r3"
    assert result is None, assertmsg

    input_dict = {
        "10.1.4.2/32": [
            {
                "prefix": "10.1.4.2/32",
                "prefixLen": 32,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.4.2/32 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.4.2/32 not installed on router r3"
    assert result is None, assertmsg

    input_dict = {
        "10.1.7.0/24": [
            {
                "prefix": "10.1.7.0/24",
                "prefixLen": 24,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.7.0/24 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.7.0/24 not installed on router r3"
    assert result is None, assertmsg

    input_dict = {}
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.8.0/24 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.8.0/24 installed on router r3"
    assert result is None, assertmsg


def test_broadcast_stub_suppression():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("Skipped because of router(s) failure")

    step("Configure R1 interface r1-eth4 with prefix suppression")
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth4\nip ospf prefix-suppression")

    step("Verify the R1 configuration of 'ip ospf prefix-suppression'")
    prefix_suppression_cfg = (
        tgen.net["r1"]
        .cmd('vtysh -c "show running ospfd" | grep "^ ip ospf prefix-suppression"')
        .rstrip()
    )
    assertmsg = "'ip ospf prefix-suppression' applied, but not present in configuration"
    assert prefix_suppression_cfg == " ip ospf prefix-suppression", assertmsg

    step("Verify that ospf-prefix suppression is applied to the R1 interface")
    r1_eth4_with_prefix_suppression = {
        "interfaces": {
            "r1-eth4": {
                "ifUp": True,
                "ospfEnabled": True,
                "ipAddress": "10.1.7.1",
                "ospfIfType": "Broadcast",
                "prefixSuppression": True,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        r1,
        "show ip ospf interface r1-eth4 json",
        r1_eth4_with_prefix_suppression,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "R1 OSPF interface r1-eth4 doesn't have prefix-suppression enabled"
    assert result is None, assertmsg

    step(
        "Verify that ospf-prefix suppression is applied to the R1 interface (non-JSON)"
    )
    prefix_suppression_show = (
        tgen.net["r1"]
        .cmd(
            'vtysh -c "show ip ospf interface r1-eth4" | grep "^  Suppress advertisement of interface IP prefix"'
        )
        .rstrip()
    )
    assertmsg = (
        "'ip ospf prefix-suppression' applied, but not present in interface show"
    )
    assert (
        prefix_suppression_show == "  Suppress advertisement of interface IP prefix"
    ), assertmsg

    step("Verify the ospf prefix is not advertised and not present on r3")
    r3 = tgen.gears["r3"]
    input_dict = {}
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.7.0/24 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.7.0/24 installed on router r3"
    assert result is None, assertmsg

    step("Remove R1 interface r1-eth4 prefix-suppression configuration")
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth4\nno ip ospf prefix-suppression")

    step("Verify no R1 configuration of 'ip ospf prefix-suppression")
    rc, _, _ = tgen.net["r1"].cmd_status(
        "show running ospfd | grep -q 'ip ospf prefix-suppression'", warn=False
    )
    assertmsg = (
        "'ip ospf prefix-suppression' not applied, but present in R1 configuration"
    )
    assert rc, assertmsg

    step("Verify that ospf-prefix suppression is not applied to the R1 interface")
    r1_eth4_without_prefix_suppression = {
        "interfaces": {
            "r1-eth4": {
                "ifUp": True,
                "ospfEnabled": True,
                "ipAddress": "10.1.7.1",
                "ospfIfType": "Broadcast",
                "prefixSuppression": False,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        r1,
        "show ip ospf interface r1-eth4 json",
        r1_eth4_without_prefix_suppression,
    )

    step("Verify that 10.1.7.0/24 route is now installed on R3")
    input_dict = {
        "10.1.7.0/24": [
            {
                "prefix": "10.1.7.0/24",
                "prefixLen": 24,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.7.0/24 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.7.0/24 not installed on router r3"
    assert result is None, assertmsg


def test_broadcast_transit_suppression():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("Skipped because of router(s) failure")

    step(
        "Configure R1 interface r1-eth0 with prefix suppression using interface address"
    )
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth0\nip ospf prefix-suppression 10.1.1.1")

    step("Verify the R1 configuration of 'ip ospf prefix-suppression 10.1.1.1'")
    prefix_suppression_cfg = (
        tgen.net["r1"]
        .cmd(
            'vtysh -c "show running ospfd" | grep "^ ip ospf prefix-suppression 10.1.1.1"'
        )
        .rstrip()
    )
    assertmsg = "'ip ospf prefix-suppression 10.1.1.1' applied, but not present in configuration"
    assert prefix_suppression_cfg == " ip ospf prefix-suppression 10.1.1.1", assertmsg

    step(
        "Configure R2 interface r2-eth0 with prefix suppression using interface address"
    )
    r2 = tgen.gears["r2"]
    r2.vtysh_cmd("conf t\ninterface r2-eth0\nip ospf prefix-suppression 10.1.1.2")

    step("Verify that ospf-prefix suppression is applied to the R1 interface")
    r1_eth0_with_prefix_suppression = {
        "interfaces": {
            "r1-eth0": {
                "ifUp": True,
                "ospfEnabled": True,
                "ipAddress": "10.1.1.1",
                "ospfIfType": "Broadcast",
                "networkType": "BROADCAST",
                "prefixSuppression": True,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        r1,
        "show ip ospf interface r1-eth0 json",
        r1_eth0_with_prefix_suppression,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "R1 OSPF interface r1-eth0 doesn't have prefix-suppression enabled"
    assert result is None, assertmsg

    step("Verify the OSPF prefix is not advertised and not present on r3")
    r3 = tgen.gears["r3"]
    input_dict = {}
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.1.0/24 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.1.0/24 installed on router r3"
    assert result is None, assertmsg

    step("Verify the OSPF Network-LSA prefixes are also not present on R3 ")
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.1.1/24 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.1.1/24 installed on router r3"
    assert result is None, assertmsg
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.1.2/24 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.1.2/24 installed on router r3"
    assert result is None, assertmsg

    step(
        "Remove R1 interface r1-eth0 prefix-suppression configuration using interface address"
    )
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth0\nno ip ospf prefix-suppression 10.1.1.1")

    step(
        "Remove R2 interface r2-eth0 prefix-suppression configuration using interface address"
    )
    r2 = tgen.gears["r2"]
    r2.vtysh_cmd("conf t\ninterface r2-eth0\nno ip ospf prefix-suppression 10.1.1.2")

    step("Verify no R1 configuration of 'ip ospf prefix-suppression")
    rc, _, _ = tgen.net["r1"].cmd_status(
        "show running ospfd | grep -q 'ip ospf prefix-suppression 10.1.1.1'", warn=False
    )
    assertmsg = "'ip ospf prefix-suppression 10.1.1.1' not applied, but present in R1 configuration"
    assert rc, assertmsg

    step("Verify that ospf-prefix suppression is not applied to the R1 interface")
    r1_eth0_without_prefix_suppression = {
        "interfaces": {
            "r1-eth0": {
                "ifUp": True,
                "ospfEnabled": True,
                "ipAddress": "10.1.1.1",
                "ospfIfType": "Broadcast",
                "networkType": "BROADCAST",
                "prefixSuppression": False,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        r1,
        "show ip ospf interface r1-eth0 json",
        r1_eth0_without_prefix_suppression,
    )

    step("Verify that 10.1.1.0/24 route is now installed on R3")
    input_dict = {
        "10.1.1.0/24": [
            {
                "prefix": "10.1.1.0/24",
                "prefixLen": 24,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.1.0/24 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.1.0/24 not installed on router r3"
    assert result is None, assertmsg


def test_nbma_transit_suppression():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("Skipped because of router(s) failure")

    step("Configure R1 interface r1-eth1 with prefix suppression")
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth1\nip ospf prefix-suppression")

    step("Configure R2 interface r2-eth1 with prefix suppression")
    r2 = tgen.gears["r2"]
    r2.vtysh_cmd("conf t\ninterface r2-eth1\nip ospf prefix-suppression")

    step("Verify that ospf-prefix suppression is applied to the R1 interface")
    r1_eth1_with_prefix_suppression = {
        "interfaces": {
            "r1-eth1": {
                "ifUp": True,
                "ospfEnabled": True,
                "ipAddress": "10.1.2.1",
                "ospfIfType": "Broadcast",
                "networkType": "NBMA",
                "prefixSuppression": True,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        r1,
        "show ip ospf interface r1-eth1 json",
        r1_eth1_with_prefix_suppression,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "R1 OSPF interface r1-eth1 doesn't have prefix-suppression enabled"
    assert result is None, assertmsg

    step("Verify the OSPF prefix is not advertised and not present on r3")
    r3 = tgen.gears["r3"]
    input_dict = {}
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.2.0/24 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.2.0/24 installed on router r3"
    assert result is None, assertmsg

    step("Verify the OSPF Network-LSA prefixes are also not present on R3 ")
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.2.1/24 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.2.1/24 installed on router r3"
    assert result is None, assertmsg
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.2.2/24 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.2.2/24 installed on router r3"
    assert result is None, assertmsg

    step("Remove R1 interface r1-eth1 prefix-suppression configuration")
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth1\nno ip ospf prefix-suppression")

    step("Remove R2 interface eth1 prefix-suppression configuration")
    r2 = tgen.gears["r2"]
    r2.vtysh_cmd("conf t\ninterface r2-eth1\nno ip ospf prefix-suppression")

    step("Verify no R1 configuration of 'ip ospf prefix-suppression")
    rc, _, _ = tgen.net["r1"].cmd_status(
        "show running ospfd | grep -q 'ip ospf prefix-suppression'", warn=False
    )
    assertmsg = (
        "'ip ospf prefix-suppression' not applied, but present in R1 configuration"
    )
    assert rc, assertmsg

    step("Verify that ospf-prefix suppression is not applied to the R1 interface")
    r1_eth1_without_prefix_suppression = {
        "interfaces": {
            "r1-eth1": {
                "ifUp": True,
                "ospfEnabled": True,
                "ipAddress": "10.1.2.1",
                "ospfIfType": "Broadcast",
                "networkType": "NBMA",
                "prefixSuppression": False,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        r1,
        "show ip ospf interface r1-eth1 json",
        r1_eth1_without_prefix_suppression,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "Prefix suppression on interface r1-eth1"
    assert result is None, assertmsg

    step("Verify that 10.1.2.0/24 route is now installed on R3")
    input_dict = {
        "10.1.2.0/24": [
            {
                "prefix": "10.1.2.0/24",
                "prefixLen": 24,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.2.0/24 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.2.0/24 not installed on router r3"
    assert result is None, assertmsg


def test_p2p_suppression():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("Skipped because of router(s) failure")

    step(
        "Configure R1 interface r1-eth2 with prefix suppression with interface address"
    )
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth2\nip ospf prefix-suppression 10.1.3.1")

    step(
        "Configure R2 interface r2-eth1 with prefix suppression with interface address"
    )
    r2 = tgen.gears["r2"]
    r2.vtysh_cmd("conf t\ninterface r2-eth2\nip ospf prefix-suppression 10.1.3.2")

    step("Verify the R1 configuration of 'ip ospf prefix-suppression 10.1.3.1'")
    prefix_suppression_cfg = (
        tgen.net["r1"]
        .cmd(
            'vtysh -c "show running ospfd" | grep "^ ip ospf prefix-suppression 10.1.3.1"'
        )
        .rstrip()
    )
    assertmsg = "'ip ospf prefix-suppression 10.1.3.1' applied, but not present in configuration"
    assert prefix_suppression_cfg == " ip ospf prefix-suppression 10.1.3.1", assertmsg

    step("Verify that ospf-prefix suppression is applied to the R1 interface")
    r1_eth2_with_prefix_suppression = {
        "interfaces": {
            "r1-eth2": {
                "ifUp": True,
                "ospfEnabled": True,
                "ipAddress": "10.1.3.1",
                "ospfIfType": "Broadcast",
                "networkType": "POINTOPOINT",
                "prefixSuppression": True,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        r1,
        "show ip ospf interface r1-eth2 json",
        r1_eth2_with_prefix_suppression,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "R1 OSPF interface r1-eth2 doesn't have prefix-suppression enabled"
    assert result is None, assertmsg

    step("Verify the OSPF prefix is not advertised and not present on r3")
    r3 = tgen.gears["r3"]
    input_dict = {}
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.3.0/24 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.3.0/24 installed on router r3"
    assert result is None, assertmsg

    step(
        "Remove R1 interface r1-eth2 prefix-suppression configuration using interface address"
    )
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth2\nno ip ospf prefix-suppression 10.1.3.1")

    step(
        "Remove R2 interface r2-eth2 prefix-suppression configuration using interface address"
    )
    r2 = tgen.gears["r2"]
    r2.vtysh_cmd("conf t\ninterface r2-eth2\nno ip ospf prefix-suppression 10.1.3.2")

    step("Verify no R1 configuration of 'ip ospf prefix-suppression")
    rc, _, _ = tgen.net["r1"].cmd_status(
        "show running ospfd | grep -q 'ip ospf prefix-suppression 10.1.3.1'", warn=False
    )
    assertmsg = "'ip ospf prefix-suppressio 10.1.3.1' not applied, but present in R1 configuration"
    assert rc, assertmsg

    step("Verify that ospf-prefix suppression is not applied to the R1 interface")
    r1_eth2_without_prefix_suppression = {
        "interfaces": {
            "r1-eth2": {
                "ifUp": True,
                "ospfEnabled": True,
                "ipAddress": "10.1.3.1",
                "ospfIfType": "Broadcast",
                "networkType": "POINTOPOINT",
                "prefixSuppression": False,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        r1,
        "show ip ospf interface r1-eth2 json",
        r1_eth2_without_prefix_suppression,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "Prefix suppression on interface r1-eth2"
    assert result is None, assertmsg

    step("Verify that 10.1.3.0/24 route is now installed on R3")
    input_dict = {
        "10.1.3.0/24": [
            {
                "prefix": "10.1.3.0/24",
                "prefixLen": 24,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.3.0/24 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.3.0/24 not installed on router r3"
    assert result is None, assertmsg


def test_p2mp_suppression():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip("Skipped because of router(s) failure")

    step("Configure R1 interface r1-eth3 with prefix suppression")
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth3\nip ospf prefix-suppression")

    step("Configure R2 interface r2-eth3 with prefix suppression")
    r2 = tgen.gears["r2"]
    r2.vtysh_cmd("conf t\ninterface r2-eth3\nip ospf prefix-suppression")

    step("Verify that ospf-prefix suppression is applied to the R1 interface")
    r1_eth3_with_prefix_suppression = {
        "interfaces": {
            "r1-eth3": {
                "ifUp": True,
                "ospfEnabled": True,
                "ipAddress": "10.1.4.1",
                "ospfIfType": "Broadcast",
                "networkType": "POINTOMULTIPOINT",
                "prefixSuppression": True,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        r1,
        "show ip ospf interface r1-eth3 json",
        r1_eth3_with_prefix_suppression,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "R1 OSPF interface r1-eth3 doesn't have prefix-suppression enabled"
    assert result is None, assertmsg

    step("Verify the OSPF P2MP prefixes are not advertised and not present on r3")
    r3 = tgen.gears["r3"]
    input_dict = {}
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.4.1/32 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.4.1/32 installed on router r3"
    assert result is None, assertmsg

    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.4.2/32 json", input_dict, True
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.4.2/32 installed on router r3"
    assert result is None, assertmsg

    step("Remove R1 interface r1-eth3 prefix-suppression configuration")
    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf t\ninterface r1-eth3\nno ip ospf prefix-suppression")

    step("Remove R2 interface r2-eth3 prefix-suppression configuration")
    r2 = tgen.gears["r2"]
    r2.vtysh_cmd("conf t\ninterface r2-eth3\nno ip ospf prefix-suppression")

    step("Verify no R1 configuration of 'ip ospf prefix-suppression")
    rc, _, _ = tgen.net["r1"].cmd_status(
        "show running ospfd | grep -q 'ip ospf prefix-suppression'", warn=False
    )
    assertmsg = (
        "'ip ospf prefix-suppression' not applied, but present in R1 configuration"
    )
    assert rc, assertmsg

    step("Verify that ospf-prefix suppression is not applied to the R1 interface")
    r1_eth3_without_prefix_suppression = {
        "interfaces": {
            "r1-eth3": {
                "ifUp": True,
                "ospfEnabled": True,
                "ipAddress": "10.1.4.1",
                "ospfIfType": "Broadcast",
                "networkType": "POINTOMULTIPOINT",
                "prefixSuppression": False,
            }
        }
    }
    test_func = partial(
        topotest.router_json_cmp,
        r1,
        "show ip ospf interface r1-eth3 json",
        r1_eth3_without_prefix_suppression,
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "Prefix suppression on interface r1-eth3"
    assert result is None, assertmsg

    step("Verify that 10.1.4.1/32 route is now installed on R3")
    input_dict = {
        "10.1.4.1/32": [
            {
                "prefix": "10.1.4.1/32",
                "prefixLen": 32,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.4.1/32 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.4.1/32 not installed on router r3"
    assert result is None, assertmsg

    step("Verify that 10.1.4.2/32 route is now installed on R3")
    input_dict = {
        "10.1.4.2/32": [
            {
                "prefix": "10.1.4.2/32",
                "prefixLen": 32,
                "protocol": "ospf",
                "nexthops": [
                    {
                        "ip": "10.1.5.2",
                        "interfaceName": "r3-eth0",
                    }
                ],
            }
        ]
    }
    test_func = partial(
        topotest.router_json_cmp, r3, "show ip route 10.1.4.2/32 json", input_dict
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)
    assertmsg = "10.1.4.2/32 not installed on router r3"
    assert result is None, assertmsg


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
